from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import functools
import math
import numpy as np

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def run(act_shape=(1, 7, 7, 15),
        weight1_shape=(7, 3, 3, 15), bias1_shape=None, scale1_shape=None,
        weight2_shape=(9, 3, 3, 7), bias2_shape=None, scale2_shape=None,
        weight3_shape=(9, 3, 3, 9), bias3_shape=None, scale3_shape=None,
        act_dtype=ng.int32,
        weight1_dtype=ng.int32, bias1_dtype=ng.int32, scale1_dtype=ng.int32,
        weight2_dtype=ng.int32, bias2_dtype=ng.int32, scale2_dtype=ng.int32,
        weight3_dtype=ng.int32, bias3_dtype=ng.int32, scale3_dtype=ng.int32,
        tmp_dtype=ng.int32,
        out_dtype=ng.int32,
        stride1=(1, 1, 1, 1), stride2=(1, 1, 1, 1), stride3=(1, 1, 1, 1),
        rshift_mul1=None, rshift_sum1=None, rshift_out1=None,
        rshift_mul2=None, rshift_sum2=None, rshift_out2=None,
        rshift_mul3=None, rshift_sum3=None, rshift_out3=None,
        act_func1=None, act_func2=None, act_func3=None,
        par_ich1=1, par_och1=1, par_col1=1, par_row1=1,
        concur_och1=None, stationary1='filter',
        par_ich2=1, par_och2=1, par_col2=1, par_row2=1,
        concur_och2=None, stationary2='filter',
        par_ich3=1, par_och3=1, par_col3=1, par_row3=1,
        concur_och3=None, stationary3='filter',
        input_ram_size1=None, filter_ram_size1=None,
        bias_ram_size1=None, scale_ram_size1=None,
        out_ram_size1=None,
        input_ram_size2=None, filter_ram_size2=None,
        bias_ram_size2=None, scale_ram_size2=None,
        out_ram_size2=None,
        input_ram_size3=None, filter_ram_size3=None,
        bias_ram_size3=None, scale_ram_size3=None,
        out_ram_size3=None,
        axi_datawidth=32, silent=False,
        filename=None, simtype='iverilog', outputfile=None):

    # create target hardware
    act = ng.placeholder(act_dtype, shape=act_shape, name='act')

    weight1 = ng.variable(weight1_dtype, shape=weight1_shape,
                          name='weight1')

    if bias1_shape is not None:
        bias1 = ng.variable(bias1_dtype, bias1_shape, name='bias1')
    else:
        bias1 = None

    if scale1_shape is not None:
        scale1 = ng.variable(scale1_dtype, scale1_shape, name='scale1')
    else:
        scale1 = None

    weight2 = ng.variable(weight2_dtype, shape=weight2_shape,
                          name='weight2')

    if bias2_shape is not None:
        bias2 = ng.variable(bias2_dtype, bias2_shape, name='bias2')
    else:
        bias2 = None

    if scale2_shape is not None:
        scale2 = ng.variable(scale2_dtype, scale2_shape, name='scale2')
    else:
        scale2 = None

    weight3 = ng.variable(weight3_dtype, shape=weight3_shape,
                          name='weight3')

    if bias3_shape is not None:
        bias3 = ng.variable(bias3_dtype, bias3_shape, name='bias3')
    else:
        bias3 = None

    if scale3_shape is not None:
        scale3 = ng.variable(scale3_dtype, scale3_shape, name='scale3')
    else:
        scale3 = None

    tmp = ng.conv2d(act, weight1, stride1,
                    bias1, scale1,
                    rshift_mul1, rshift_sum1, rshift_out1,
                    act_func1, 'SAME', False,
                    tmp_dtype, ng.int32, ng.int32,
                    'conv2d_1',
                    par_ich1, par_och1, par_col1, par_row1,
                    concur_och1, stationary1,
                    input_ram_size1, filter_ram_size1,
                    bias_ram_size1, scale_ram_size1,
                    None, None, None,
                    out_ram_size1)

    tmp = ng.conv2d(tmp, weight2, stride2,
                    bias2, scale2,
                    rshift_mul2, rshift_sum2, rshift_out2,
                    act_func2, 'SAME', False,
                    out_dtype, ng.int32, ng.int32,
                    'conv2d_2',
                    par_ich2, par_och2, par_col2, par_row2,
                    concur_och2, stationary2,
                    input_ram_size2, filter_ram_size2,
                    bias_ram_size2, scale_ram_size2,
                    None, None, None,
                    out_ram_size2)

    out = ng.conv2d(tmp, weight3, stride3,
                    bias3, scale3,
                    rshift_mul3, rshift_sum3, rshift_out3,
                    act_func3, 'SAME', False,
                    out_dtype, ng.int32, ng.int32,
                    'conv2d_3',
                    par_ich3, par_och3, par_col3, par_row3,
                    concur_och3, stationary3,
                    input_ram_size3, filter_ram_size3,
                    bias_ram_size3, scale_ram_size3,
                    None, None, None,
                    out_ram_size3)

    targ = ng.to_veriloggen([out], 'matrix_conv2d_conv2d_conv2d', silent=silent,
                            config={'maxi_datawidth': axi_datawidth})

    # verification data
    vact = np.arange(act.length, dtype=np.int64).reshape(act.shape) % [16]

    vweight1 = np.arange(weight1.length,
                         dtype=np.int64).reshape(weight1_shape) % [32] - [16]

    if bias1 is not None:
        vbias1 = np.arange(bias1.length,
                           dtype=np.int64).reshape(bias1.shape) % [16]
    else:
        vbias1 = None

    if scale1 is not None:
        vscale1 = np.arange(scale1.length,
                            dtype=np.int64).reshape(scale1.shape) % [8]
    else:
        vscale1 = None

    vweight2 = np.arange(weight2.length,
                         dtype=np.int64).reshape(weight2_shape) % [32] - [16]

    if bias2 is not None:
        vbias2 = np.arange(bias2.length,
                           dtype=np.int64).reshape(bias2.shape) % [16]
    else:
        vbias2 = None

    if scale2 is not None:
        vscale2 = np.arange(scale2.length,
                            dtype=np.int64).reshape(scale2.shape) % [8]
    else:
        vscale2 = None

    vweight3 = np.arange(weight3.length,
                         dtype=np.int64).reshape(weight3_shape) % [32] - [16]

    if bias3 is not None:
        vbias3 = np.arange(bias3.length,
                           dtype=np.int64).reshape(bias3.shape) % [16]
    else:
        vbias3 = None

    if scale3 is not None:
        vscale3 = np.arange(scale3.length,
                            dtype=np.int64).reshape(scale3.shape) % [8]
    else:
        vscale3 = None

    eval_outs = ng.eval([out], act=vact, weight1=vweight1, bias1=vbias1, scale1=vscale1,
                        weight2=vweight2, bias2=vbias2, scale2=vscale2,
                        weight3=vweight3, bias3=vbias3, scale3=vscale3)
    vout = eval_outs[0]

    # to memory image
    size_max = int(math.ceil(max(act.memory_size, weight1.memory_size,
                                 bias1.memory_size if bias1 is not None else 0,
                                 scale1.memory_size if scale1 is not None else 0,
                                 weight2.memory_size,
                                 bias2.memory_size if bias2 is not None else 0,
                                 scale2.memory_size if scale2 is not None else 0,
                                 weight3.memory_size,
                                 bias3.memory_size if bias3 is not None else 0,
                                 scale3.memory_size if scale3 is not None else 0,
                                 out.memory_size) / 4096)) * 4096
    check_addr = max(act.addr, weight1.addr,
                     bias1.addr if bias1 is not None else -1,
                     scale1.addr if scale1 is not None else -1,
                     weight2.addr,
                     bias2.addr if bias2 is not None else -1,
                     scale2.addr if scale2 is not None else -1,
                     weight3.addr,
                     bias3.addr if bias3 is not None else -1,
                     scale3.addr if scale3 is not None else -1,
                     out.addr) + size_max
    size_check = size_max
    tmp_addr = check_addr + size_check

    memimg_datawidth = 32
    mem = np.zeros([1024 * 1024 * 8 // (memimg_datawidth // 8)], dtype=np.int64)
    mem = mem + [100]

    axi.set_memory(mem, vact, memimg_datawidth,
                   act_dtype.width, act.addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), par_ich1))

    axi.set_memory(mem, vweight1, memimg_datawidth,
                   weight1_dtype.width, weight1.addr,
                   max(int(math.ceil(axi_datawidth / weight1_dtype.width)), par_ich1))
    if bias1_shape is not None:
        axi.set_memory(mem, vbias1, memimg_datawidth,
                       bias1_dtype.width, bias1.addr,
                       max(int(math.ceil(axi_datawidth / bias1_dtype.width)), par_och1))
    if scale1_shape is not None:
        axi.set_memory(mem, vscale1, memimg_datawidth,
                       scale1_dtype.width, scale1.addr,
                       max(int(math.ceil(axi_datawidth / scale1_dtype.width)), par_och1))

    axi.set_memory(mem, vweight2, memimg_datawidth,
                   weight2_dtype.width, weight2.addr,
                   max(int(math.ceil(axi_datawidth / weight2_dtype.width)), par_ich2))
    if bias2_shape is not None:
        axi.set_memory(mem, vbias2, memimg_datawidth,
                       bias2_dtype.width, bias2.addr,
                       max(int(math.ceil(axi_datawidth / bias2_dtype.width)), par_och2))
    if scale2_shape is not None:
        axi.set_memory(mem, vscale2, memimg_datawidth,
                       scale2_dtype.width, scale2.addr,
                       max(int(math.ceil(axi_datawidth / scale2_dtype.width)), par_och2))

    axi.set_memory(mem, vweight3, memimg_datawidth,
                   weight3_dtype.width, weight3.addr,
                   max(int(math.ceil(axi_datawidth / weight3_dtype.width)), par_ich3))
    if bias3_shape is not None:
        axi.set_memory(mem, vbias3, memimg_datawidth,
                       bias3_dtype.width, bias3.addr,
                       max(int(math.ceil(axi_datawidth / bias3_dtype.width)), par_och3))
    if scale3_shape is not None:
        axi.set_memory(mem, vscale3, memimg_datawidth,
                       scale3_dtype.width, scale3.addr,
                       max(int(math.ceil(axi_datawidth / scale3_dtype.width)), par_och3))

    axi.set_memory(mem, vout, memimg_datawidth,
                   out_dtype.width, check_addr,
                   max(int(math.ceil(axi_datawidth / out_dtype.width)), par_och3))

    # test controller
    m = Module('test')
    params = m.copy_params(targ)
    ports = m.copy_sim_ports(targ)
    clk = ports['CLK']
    resetn = ports['RESETN']
    rst = m.Wire('RST')
    rst.assign(Not(resetn))

    # AXI memory model
    if outputfile is None:
        outputfile = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + outputfile

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst,
                                datawidth=axi_datawidth,
                                memimg=mem, memimg_name=memimg_name,
                                memimg_datawidth=memimg_datawidth)
    memory.connect(ports, 'maxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    # timer
    time_counter = m.Reg('time_counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        time_counter.inc()
    )

    def ctrl():
        for i in range(100):
            pass

        ng.sim.set_global_addrs(_saxi, tmp_addr)

        start_time = time_counter.value
        ng.sim.start(_saxi)

        print('# start')

        ng.sim.wait(_saxi)
        end_time = time_counter.value

        print('# end')
        print('# execution cycles: %d' % (end_time - start_time))

        # verify
        ok = True
        for bat in range(out.shape[0]):
            for y in range(out.shape[1]):
                for x in range(out.shape[2]):
                    for ch in range(out.shape[3]):
                        orig = memory.read_word(
                            bat * out.aligned_shape[1] * out.aligned_shape[2] * out.aligned_shape[3] +
                            y * out.aligned_shape[2] * out.aligned_shape[3] +
                            x * out.aligned_shape[3] + ch,
                            out.addr, out_dtype.width)
                        check = memory.read_word(
                            bat * out.aligned_shape[1] * out.aligned_shape[2] * out.aligned_shape[3] +
                            y * out.aligned_shape[2] * out.aligned_shape[3] +
                            x * out.aligned_shape[3] + ch,
                            check_addr, out_dtype.width)
                        if vthread.verilog.NotEql(orig, check):
                            print('NG (', bat, y, x, ch,
                                  ') orig: ', orig, ' check: ', check)
                            ok = False
                        # else:
                        #    print('OK (', bat, y, x, ch,
                        #          ') orig: ', orig, ' check: ', check)

        if ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(targ, 'uut',
                     params=m.connect_params(targ),
                     ports=m.connect_ports(targ))

    # simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, resetn, m.make_reset(), period=100, polarity='low')

    init.add(
        Delay(10000000),
        Systask('finish'),
    )

    # output source code
    if filename is not None:
        m.to_verilog(filename)

    # run simulation
    sim = simulation.Simulator(m, sim=simtype)
    rslt = sim.run(outputfile=outputfile)

    return rslt


if __name__ == '__main__':
    rslt = run(silent=False, filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import functools
from collections import OrderedDict

import nngen.basic_types as bt


class relu(bt._ActFuncOperator):
    """
    Applies the rectified linear unit function element-wise:

    .. math::
        {ReLU}(x) = \max(0, x)

    Args:
        features: A Tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Mux(args[0] > strm.Int(0), args[0], strm.Int(0))

    def __init__(self, features, dtype=None, name=None, par=1):
        shape = None
        bt._ActFuncOperator.__init__(self, features,
                                     dtype=dtype, shape=shape, name=name, par=par)

    def get_eval_method(self):
        import nngen.verify as verify

        name = self.__class__.__name__
        method = getattr(verify, name, None)
        method = functools.partial(method,
                                   features_dtype=self.args[0].dtype)
        return method


class relu6(relu):
    """
    Applies the element-wise function:

    .. math::
        {ReLU6}(x) = \min(\max(0, x), 6)

    Args:
        features: A Tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    def get_local_control_param_values(self):
        return OrderedDict([('max_val', round(self.args[0].scale_factor) * 6)])

    def op(self, strm, *args, **kwargs):
        return strm.Mux(args[0] > strm.Int(0),
                        strm.Mux(args[0] > self.max_val, self.max_val, args[0]),
                        strm.Int(0))

    def get_eval_method(self):
        import nngen.verify as verify

        name = self.__class__.__name__
        method = getattr(verify, name, None)
        method = functools.partial(method,
                                   features_dtype=self.args[0].dtype,
                                   features_scale_factor=round(self.args[0].scale_factor))
        return method



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def upsampling2d(value, factors,
                 dtype=None, name=None, par=1,
                 value_dtype=None):

    ret_shape = [s * f for s, f in zip(value.shape, factors)]
    ret = np.zeros(ret_shape, dtype=np.int64)

    factor_col = factors[2]
    factor_row = factors[1]

    for bat in range(ret_shape[0]):
        for row in range(ret_shape[1]):
            for col in range(ret_shape[2]):
                ret[bat][row][col][:] = value[bat][row // factor_row][col // factor_col]

    return ret



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
import functools
import numpy as np

import veriloggen.types.axi as axi

from . import basic_types as bt
from . import storage as st


def to_axis(axis, rank):
    if axis is None:
        return None

    if not isinstance(axis, (tuple, list)):
        axis = (axis,)

    new_axis = []
    for i in axis:
        if i < 0:
            i += rank
        new_axis.append(i)

    return tuple(new_axis)


def to_reduce_shape(input_shape, axis=None, keep_dims=False):
    if not keep_dims:
        if axis is None:
            shape = tuple([1])

        else:
            shape = []
            for i, s in enumerate(input_shape):
                if i not in axis:
                    shape.append(s)

            if len(shape) == 0:
                shape.append(1)

            shape = tuple(shape)

    else:
        if axis is None:
            shape = []
            for s in input_shape:
                shape.append(1)

            shape = tuple(shape)

        else:
            shape = []
            for i, s in enumerate(input_shape):
                if i not in axis:
                    shape.append(s)
                else:
                    shape.append(1)

            shape = tuple(shape)

    return shape


def pix_size(size, ksize, stride, padding="SAME"):
    if padding == "SAME":
        return int(math.ceil(size / stride))
    elif padding == "VALID":
        return int(math.ceil((size - ksize + 1) / stride))
    raise ValueError("not supported padding type: '%s'" % padding)


def pad_size(size, ksize, stride):
    if size % stride == 0:
        return max(ksize - stride, 0)
    return max(ksize - (size % stride), 0)


def pad_size_split(size, ksize, stride):
    pad = pad_size(size, ksize, stride)
    a = pad // 2
    b = pad - a
    return pad, a, b


def to_storage_dict(*args, **kwargs):
    d = {}

    for key, arg in kwargs.items():
        if not bt.is_storage(arg) and not arg.is_output:
            raise ValueError("'%s' is not input/output storage." % str(arg))

        while bt.is_view(arg) or bt.is_removable_reshape(arg):
            arg = arg.args[0]

        d[key] = arg

    for arg in args:
        if not bt.is_storage(arg) and not arg.is_output:
            raise ValueError("'%s' is not input/output storage." % str(arg))

        while bt.is_view(arg) or bt.is_removable_reshape(arg):
            arg = arg.args[0]

        d[arg.name] = arg

    return d


def export_ndarray(objs, chunk_size=64):

    if not isinstance(objs, (list, tuple)):
        objs = [objs]

    objs = _collect_numerics(objs)

    variables = []
    for obj in objs:
        if isinstance(obj, st.variable):
            variables.append(obj)

    constants = []
    for obj in objs:
        if isinstance(obj, st.constant):
            constants.append(obj)

    return make_ndarray(variables, constants, chunk_size)


def _collect_numerics(objs):
    ret = set()
    for obj in objs:
        if obj not in ret:
            ret.update(obj._collect_numerics())

    ret = sorted(list(ret), key=lambda x: x.object_id)
    return ret


def make_ndarray(variables, constants, chunk_size=64):
    min_addr = 2 ** 64
    max_addr = 0

    if isinstance(variables, dict):
        variables = list(variables.values())

    if isinstance(constants, dict):
        constants = list(constants.values())

    for variable in variables:
        if not isinstance(variable, st.variable):
            raise TypeError("'%s' is not variable.'" % str(variable))

        if variable.maxi is None:
            continue

        min_addr = min(min_addr, variable.addr)
        size = aligned_size(variable.memory_size, chunk_size)
        max_addr = max(max_addr, variable.addr + size)

    for constant in constants:
        if not isinstance(constant, st.constant):
            raise TypeError("'%s' is not _Constant.'" % str(constant))

        if constant.maxi is None:
            continue

        min_addr = min(min_addr, constant.addr)
        size = aligned_size(constant.memory_size, chunk_size)
        max_addr = max(max_addr, constant.addr + size)

    # return empty
    if max_addr < min_addr:
        param = np.zeros([0], dtype=np.uint8)
        return param

    param = np.zeros([max_addr - min_addr], dtype=np.uint8)
    dst_width = 8

    for variable in variables:
        if variable.maxi is None:
            continue
        if variable.value is None:
            continue

        src_width = variable.dtype.width
        dst_offset = variable.addr - min_addr
        alignment = variable.get_word_alignment()
        axi.set_memory(param, variable.value, dst_width, src_width, dst_offset, alignment)

    for constant in constants:
        if constant.maxi is None:
            continue

        src_width = constant.dtype.width
        dst_offset = constant.addr - min_addr
        alignment = constant.get_word_alignment()
        axi.set_memory(param, constant.value, dst_width, src_width, dst_offset, alignment)

    return param


def aligned_size(size, chunk_size):
    return int(math.ceil(size / chunk_size)) * chunk_size


def is_elementwise_operator(obj):
    return isinstance(obj, bt._ElementwiseOperator)


def clip_threshold(width, signed, asymmetric_clip=False):
    if signed:
        p_th = (1 << (width - 1)) - 1
        if asymmetric_clip:
            n_th = -1 * p_th - 1
        else:
            n_th = -1 * p_th
    else:
        p_th = (1 << width) - 1
        n_th = 0

    return p_th, n_th



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from . import identity


def Floor(visitor, node):

    return identity.Identity(visitor, node)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from collections import OrderedDict

import nngen.basic_types as bt
import nngen.verilog as verilog


class extern(bt._Operator):
    input_chainable = False
    output_chainable = False
    parallel_scheduling_allowed = False

    def __sub_str__(self):
        opcode = ' opcode:%d' % self.opcode
        return opcode

    def __init__(self, values, opcode,
                 shape=None, dtype=None, name=None,
                 func=None):

        if not isinstance(values, (tuple, list)):
            values = [values]

        if not isinstance(opcode, int):
            raise TypeError("opcode must be int, not '%s'." %
                            str(type(opcode)))

        if opcode <= 0:
            raise ValueError('opcode must be greater than 0.')

        if shape is None:
            shape = values[0].shape

        if dtype is None:
            dtype = values[0].dtype

        bt._Operator.__init__(self, *values,
                              dtype=dtype, shape=shape, name=name)

        self.opcode = opcode

        # for verify
        self.func = func

    def attribute(self):
        pass

    def get_required_rams(self):
        """
        @return 3 tuples of (width, length)
        """

        inputs = ()
        outputs = ()
        temps = ()
        return inputs, outputs, temps

    def get_stream_func(self):
        return None

    def get_stream_hash(self):
        clsinfo = [type(self)]
        clsinfo = tuple(clsinfo)
        return (clsinfo,)

    def get_control_param_values(self):
        opcode_value = self.opcode

        return OrderedDict([('opcode_value', opcode_value)])

    def control_sequence(self, fsm):
        # send to SW
        self.saxi.write(fsm, verilog.control_reg_extern_send,
                        self.opcode_value)
        # receive from SW
        self.saxi.wait_flag(fsm, verilog.control_reg_extern_recv, 0,
                            resetvalue=0, polarity=False)
        # reset opcode
        self.saxi.write(fsm, verilog.control_reg_extern_send, 0)

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = self.__class__.__name__

        args = [arg.eval(memo, input_dict)
                for arg in self.args]

        values = args
        opcode = self.opcode
        kwargs['shape'] = self.shape
        kwargs['dtype'] = self.dtype
        kwargs['name'] = self.name
        kwargs['value_dtypes'] = [arg.dtype for arg in self.args]
        kwargs['func'] = self.func

        method = self.get_eval_method()
        ret = method(values, opcode, **kwargs)
        memo[id(self)] = ret

        return ret



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
import veriloggen

import matrix_less


a_shape = (15, 15)
b_shape = (15, 15)
a_dtype = ng.int32
b_dtype = ng.int32
c_dtype = ng.int32
par = 1
axi_datawidth = 32


def test(request, silent=True):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = matrix_less.run(a_shape, b_shape,
                           a_dtype, b_dtype, c_dtype,
                           par, axi_datawidth, silent,
                           filename=None, simtype=simtype,
                           outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = [line for line in rslt.splitlines() if line.startswith('# verify:')][0]
    assert(verify_rslt == '# verify: PASSED')


if __name__ == '__main__':
    rslt = matrix_less.run(a_shape, b_shape,
                           a_dtype, b_dtype, c_dtype,
                           par, axi_datawidth, silent=False,
                           filename='tmp.v',
                           outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
import functools
from collections import OrderedDict

import veriloggen as vg

import nngen.basic_types as bt
import nngen.util as util


class concat(bt._Operator):
    input_chainable = False
    output_chainable = False

    def __sub_str__(self):
        axis = ' axis:%d' % self.axis
        buffered = (' buffered'
                    if hasattr(self, 'buffered_value') and self.buffered_value else '')
        return ''.join([axis, buffered])

    def __init__(self, values, axis, dtype=None, name=None):
        rank = bt.get_rank(values[0].shape)
        _dtype = values[0].dtype

        for value in values:
            r = bt.get_rank(value.shape)
            if r != rank:
                raise ValueError('all values must have a same rank: %d != %d' %
                                 (r, rank))
            rank = r
            d = value.dtype
            if d != _dtype:
                raise ValueError('all values must have a same dtype: %s != %s' %
                                 (d, _dtype))
            _dtype = d

        if isinstance(axis, (tuple, list)):
            raise TypeError('axis must be int, not tuple or list.')

        axis = util.to_axis(axis, rank)[0]

        shape = []
        for i in range(rank):
            size = 0
            for value in values:
                if i == axis:
                    size += value.shape[i]
                else:
                    if size != 0 and size != value.shape[i]:
                        raise ValueError(
                            'all values must have a same shape, excluding axis: %d != %d' %
                            (size, value.shape[i]))
                    size = max(size, value.shape[i])

            shape.append(size)

        shape = tuple(shape)

        bt._Operator.__init__(self, *values,
                              dtype=dtype, shape=shape, name=name)
        self.axis = axis

    def attribute(self):
        pass

    def get_required_rams(self):
        arg_width = 0
        arg_len = 0
        for arg in self.args:
            arg_width = max(arg_width, arg.get_ram_width())
            arg_len = max(arg_len, arg.shape[-1])

        inputs = [(arg_width, arg_len)]
        outputs = [(self.get_ram_width(), self.shape[-1])]
        temps = []
        return inputs, outputs, temps

    def get_stream_func(self):
        def func(strm):
            datawidth = self.args[0].get_op_width()
            src = strm.source(datawidth=datawidth)
            strm.sink(src)

        return func

    def get_control_param_values(self):
        buffered = False

        for arg in self.args:
            if self.dtype != arg.dtype:
                buffered = True

        if self.axis == bt.get_rank(self.shape) - 1:
            for arg in self.args:
                if arg.shape[-1] != arg.get_aligned_shape()[-1]:
                    buffered = True

        # for __str__
        self.buffered_value = buffered

        aligned_shape = self.get_aligned_shape()
        aligned_length = self.get_aligned_length()

        arg_read_sizes = [arg.shape[-1] for arg in self.args]
        arg_addr_incs = [bt.to_byte(bt.align_word(arg.shape[-1], arg.get_word_alignment()) *
                                    arg.get_ram_width())
                         for arg in self.args]

        arg_chunk_sizes = [functools.reduce(lambda x, y: x * y, arg.shape[self.axis:-1], 1)
                           for arg in self.args]

        out_write_size = aligned_shape[-1]
        out_addr_inc = bt.to_byte(bt.align_word(self.shape[-1], self.get_word_alignment()) *
                                  self.get_ram_width())

        num_steps = int(math.ceil(aligned_length / out_write_size))
        if not buffered:
            num_steps *= len(self.args)

        return OrderedDict([('buffered', buffered),
                            ('arg_read_sizes', arg_read_sizes),
                            ('arg_addr_incs', arg_addr_incs),
                            ('arg_chunk_sizes', arg_chunk_sizes),
                            ('out_write_size', out_write_size),
                            ('out_addr_inc', out_addr_inc),
                            ('num_steps', num_steps)])

    def control_sequence(self, fsm):
        arg_gaddrs = [self.m.Reg(self._name('arg_gaddr_%d' % i),
                                 self.maxi.addrwidth, initval=0)
                      for i, _ in enumerate(self.arg_objaddrs)]
        out_gaddr = self.m.Reg(self._name('out_gaddr'),
                               self.maxi.addrwidth, initval=0)

        arg_laddr = self.m.Reg(self._name('arg_laddr'),
                               self.maxi.addrwidth, initval=0)
        copy_laddr = self.m.Reg(self._name('copy_laddr'),
                                self.maxi.addrwidth, initval=0)
        copy_size = self.m.Reg(self._name('copy_size'),
                               self.maxi.addrwidth, initval=0)
        sum_read_sizes = self.m.Wire(self._name('sum_read_sizes'),
                                     max([i.width for i in self.arg_read_sizes]) +
                                     int(math.ceil(math.log2(len(self.arg_read_sizes)))))
        sum_sizes = self.arg_read_sizes[0]
        for size in self.arg_read_sizes[1:]:
            sum_sizes = vg.Add(sum_sizes, size)
        sum_read_sizes.assign(sum_sizes)

        out_addr_inc_unbuffered = self.m.Reg(self._name('out_addr_inc_unbuffered'),
                                             self.maxi.addrwidth, initval=0)

        arg_select = self.m.Reg(self._name('arg_select'),
                                int(max(math.ceil(math.log(len(self.args), 2)), 1)),
                                initval=0)
        arg_chunk_count = self.m.Reg(self._name('arg_chunk_count'),
                                     self.maxi.addrwidth + 1, initval=0)
        out_count = self.m.Reg(self._name('out_count'),
                               self.maxi.addrwidth + 1, initval=0)

        # --------------------
        # initialization phase
        # --------------------
        fsm(
            [arg_gaddr(0) for arg_gaddr in arg_gaddrs],
            out_gaddr(0),
            arg_laddr(0),
            copy_laddr(0),
            copy_size(0),
            out_addr_inc_unbuffered(0),
            arg_select(0),
            arg_chunk_count(0),
            out_count(0)
        )

        fsm.goto_next()

        # --------------------
        # Read phase
        # --------------------
        state_read = fsm.current

        fsm.inc()

        state_read_begin_list = []
        state_read_end_list = []

        for (arg,
             arg_objaddr, arg_gaddr, arg_addr_inc,
             arg_read_size, arg_chunk_size) in zip(
                 self.args,
                 self.arg_objaddrs, arg_gaddrs, self.arg_addr_incs,
                 self.arg_read_sizes, self.arg_chunk_sizes):

            b = fsm.current
            state_read_begin_list.append(b)

            # normal
            laddr = arg_laddr
            gaddr = arg_objaddr + arg_gaddr

            bt.bus_lock(self.maxi, fsm)
            bt.dma_read(self.maxi, fsm, self.input_rams[0], laddr, gaddr, arg_read_size)
            bt.bus_unlock(self.maxi, fsm)

            fsm(
                arg_gaddr.add(arg_addr_inc),
                arg_laddr.add(arg_read_size),
                arg_chunk_count.inc(),
                copy_size(arg_read_size),
                out_addr_inc_unbuffered(arg_addr_inc)
            )
            fsm.If(arg_chunk_count == arg_chunk_size - 1)(
                arg_chunk_count(0),
                arg_select.inc()
            )
            fsm.If(arg_chunk_count == arg_chunk_size - 1,
                   arg_select == len(self.args) - 1)(
                arg_select(0)
            )

            e = fsm.current
            state_read_end_list.append(e)

            fsm.inc()

        state_read_end = fsm.current

        for i, b in enumerate(state_read_begin_list):
            fsm.If(arg_select == i).goto_from(state_read, b)

        for i, e in enumerate(state_read_end_list):
            fsm.goto_from(e, state_read_end)

        # --------------------
        # Copy phase
        # --------------------
        state_copy = fsm.current

        name = list(self.stream.sources.keys())[0]
        self.stream.set_source(fsm, name, self.input_rams[0], 0, copy_size)
        fsm.set_index(fsm.current - 1)

        name = list(self.stream.sinks.keys())[0]
        self.stream.set_sink(fsm, name, self.output_rams[0], copy_laddr, copy_size)
        self.stream.run(fsm)
        self.stream.join(fsm)

        fsm(
            arg_laddr(0),
            copy_laddr.add(copy_size)
        )
        fsm.goto_next()

        fsm.If(copy_laddr < sum_read_sizes).goto(state_read)
        fsm.If(copy_laddr >= sum_read_sizes).goto_next()

        state_copy_end = fsm.current
        fsm.If(vg.Not(self.buffered)).goto_from(state_copy, state_copy_end)

        # --------------------
        # Write phase
        # --------------------
        state_write = fsm.current
        fsm.inc()

        # Case with Copy
        state_write_buffered = fsm.current

        laddr = 0
        gaddr = self.objaddr + out_gaddr
        bt.bus_lock(self.maxi, fsm)
        bt.dma_write(self.maxi, fsm,
                     self.output_rams[0], laddr, gaddr, self.out_write_size)
        bt.bus_unlock(self.maxi, fsm)

        fsm(
            copy_laddr(0),
            out_gaddr.add(self.out_addr_inc),
            out_count.inc()
        )

        state_write_end_buffered = fsm.current
        fsm.inc()

        # Case without Copy
        state_write_unbuffered = fsm.current

        laddr = 0
        gaddr = self.objaddr + out_gaddr
        bt.bus_lock(self.maxi, fsm)
        bt.dma_write(self.maxi, fsm,
                     self.input_rams[0], laddr, gaddr, copy_size)
        bt.bus_unlock(self.maxi, fsm)

        fsm(
            arg_laddr(0),
            out_gaddr.add(out_addr_inc_unbuffered),
            out_count.inc()
        )

        state_write_end_unbuffered = fsm.current
        fsm.inc()

        state_write_end = fsm.current

        fsm.If(self.buffered).goto_from(state_write, state_write_buffered)
        fsm.If(vg.Not(self.buffered)).goto_from(state_write, state_write_unbuffered)

        fsm.goto_from(state_write_end_buffered, state_write_end)
        fsm.goto_from(state_write_end_unbuffered, state_write_end)

        # --------------------
        # update for next iteration
        # --------------------
        fsm.If(out_count < self.num_steps).goto(state_read)
        fsm.If(out_count == self.num_steps).goto_next()

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = self.__class__.__name__

        values = [arg.eval(memo, input_dict)
                  for arg in self.args]

        kwargs['axis'] = self.axis
        kwargs['dtype'] = self.dtype
        kwargs['name'] = self.name

        method = self.get_eval_method()
        ret = method(values, **kwargs)
        memo[id(self)] = ret

        return ret



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
import collections
import copy
import inspect
import types

import veriloggen as vg
import veriloggen.types.axi as axi
import veriloggen.thread as vthread
import veriloggen.types.ipxact as ipxact

from . import basic_types as bt
from . import dtype_list
from . import scheduler
from . import version
from . import substreams


default_config = {
    # clock, reset
    'clock_name': 'CLK',
    'reset_name': 'RESETN',
    'reset_polarity': 'low',

    # measure enable main_fsm
    'measurable_main_fsm': True,

    # interrupt
    'interrupt_enable': True,
    'interrupt_name': 'irq',

    # AXI
    'axi_coherent': False,
    'maxi_datawidth': 32,
    'maxi_addrwidth': 32,
    'saxi_datawidth': 32,
    'saxi_addrwidth': 32,
    'req_fifo_addrwidth': 3,

    # address map
    'default_global_addr_offset': 0,
    'use_map_ram': False,  # using a address map RAM instead of AXIS registers
    'use_map_reg': False,  # using dedicated address registers instead of unified register for variable and constant

    # control params
    'use_param_ram': False,  # using a RAM instead of MUX for control params.
    'min_param_ram_len': 0,  # if use_param_ram == True and num_of_params < min_param_ram_len,
                             # a control param RAM is created instead of MUX

    # default parameters
    'default_datawidth': 32,
    'min_onchip_ram_capacity': 32 * 128,
    'offchipram_chunk_bytes': 64,
    'max_parallel_ops': 1,

    # RAM style annotation
    'onchip_ram_style': None,  # '(* ram_style = "block" *)' for Xilinx
    'param_ram_style': None,  # '(* ram_style = "block" *)' for Xilinx
    'map_ram_style': None,  # '(* ram_style = "block" *)' for Xilinx

    'onchip_ram_priority': 'total_size',
    # 'onchip_ram_priority': 'max_size',
    # 'onchip_ram_priority': (lambda cur, width, length, num: cur + width * length * num),

    # for debug
    'fsm_as_module': False,
    'disable_stream_cache': False,
    'disable_control_cache': False,
    'dump_stream': False,
    'dump_stream_base': 10,
}

max_burst_length = 256

num_header_regs = 4
header_reg = 0  # index used in "sim.py"

# control_reg_global_offset must be same as REG_GLOBAL_OFFSET in pynq/nngen_ctrl.py'.
control_reg_global_offset = 32
num_control_regs = control_reg_global_offset + 1 - num_header_regs

num_control_regs = 33 - num_header_regs
control_reg_start = num_header_regs + 0
control_reg_busy = num_header_regs + 1
control_reg_reset = num_header_regs + 2
control_reg_extern_send = num_header_regs + 3
control_reg_extern_recv = num_header_regs + 4

control_reg_interrupt_isr = num_header_regs + 5
control_reg_interrupt_isr_busy = 0  # bit-field index in ISR
control_reg_interrupt_isr_extern = 1  # bit-field index in ISR
control_reg_interrupt_ier = num_header_regs + 6
control_reg_interrupt_iar = num_header_regs + 7

control_reg_count = num_header_regs + 8
control_reg_count_state = num_header_regs + 9
control_reg_count_div = num_header_regs + 10

control_reg_reserved = num_header_regs + 11  # head of reserved region

control_reg_address_amount = control_reg_global_offset - 1
control_reg_global_addr = control_reg_global_offset + 1

# when config['use_map_ram'] is True
num_addr_map_regs = 3
control_reg_load_global_addr_map = num_header_regs + num_control_regs
control_reg_busy_global_addr_map = num_header_regs + num_control_regs + 1
control_reg_addr_global_addr_map = num_header_regs + num_control_regs + 2


def to_veriloggen(objs, name, config=None, silent=False):

    config = load_default_config(config)

    m = _to_veriloggen_module(objs, name, config, silent=silent)

    return m


def to_verilog(objs, name, filename=None, config=None, silent=False):

    config = load_default_config(config)

    m = _to_veriloggen_module(objs, name, config,
                              silent=silent, where_from='to_verilog', output=filename)

    verilog_code = m.to_verilog(filename)

    return verilog_code


def to_ipxact(objs, name, ipname=None, config=None, silent=False):

    if ipname is None:
        ipname = name

    config = load_default_config(config)

    m = _to_veriloggen_module(objs, name, config,
                              silent=silent, where_from='to_ipxact', output=ipname)

    clk_name = config['clock_name']
    rst_name = config['reset_name']
    rst_polarity = ('ACTIVE_LOW'
                    if 'low' in config['reset_polarity'] else
                    'ACTIVE_HIGH')
    irq_name = config['interrupt_name']
    irq_sensitivity = 'LEVEL_HIGH'

    clk_ports = [(clk_name, (rst_name,))]
    rst_ports = [(rst_name, rst_polarity)]
    irq_ports = [(irq_name, irq_sensitivity)] if config['interrupt_enable'] else []

    ipxact.to_ipxact(m, ipname,
                     clk_ports=clk_ports,
                     rst_ports=rst_ports,
                     irq_ports=irq_ports)

    return m


def load_default_config(config=None):
    my_config = copy.copy(default_config)

    if config is not None:
        my_config.update(config)

    config = my_config

    return config


def _to_veriloggen_module(objs, name, config=None,
                          silent=False, where_from=None, output=None):

    if not isinstance(objs, (list, tuple)):
        objs = [objs]

    (objs, num_storages,
     num_input_storages, num_output_storages) = analyze(config, objs)
    m, clk, rst, maxi, saxi = make_module(config, name, objs,
                                          num_storages, num_input_storages,
                                          num_output_storages)

    schedule_table = schedule(config, objs)

    header_info = make_header_addr_map(config, saxi)

    (ram_dict, substrm_dict, ram_set_cache,
     stream_cache, control_cache, main_fsm,
     global_map_info, global_mem_map) = allocate(config, m, clk, rst,
                                                 maxi, saxi, objs, schedule_table)

    reg_map = make_reg_map(config, global_map_info, header_info)

    saxi.register[control_reg_address_amount].initval = address_space_amount(global_mem_map)

    if not silent:
        dump_config(config, where_from, output)
        dump_schedule_table(schedule_table)
        dump_rams(ram_dict)
        dump_substreams(substrm_dict)
        dump_streams(stream_cache)
        dump_main_fsm(main_fsm)
        dump_controls(control_cache, main_fsm)
        dump_register_map(reg_map)
        dump_memory_map(global_mem_map)

    return m


def address_space_amount(mem_map):
    max_gaddr = 0
    min_gaddr = 0
    for start, end in mem_map.keys():
        if start <= end:
            max_gaddr = max(max_gaddr, start, end)
            min_gaddr = min(min_gaddr, start, end)

    num_bytes = max_gaddr - min_gaddr + 1

    return num_bytes


def analyze(config, objs):
    set_output(objs)

    objs = collect_numerics(objs)
    num_storages = count_storages(objs)
    num_input_storages = count_input_storages(objs)
    num_output_storages = count_output_storages(objs)

    set_default_dtype(config, objs)

    return objs, num_storages, num_input_storages, num_output_storages


def set_output(objs):
    for obj in objs:
        obj.set_output()


def collect_numerics(objs):
    ret = set()
    for obj in objs:
        if obj not in ret:
            ret.update(obj._collect_numerics())

    ret = sorted(list(ret), key=lambda x: x.object_id)
    return ret


def count_storages(objs):
    count = 0
    for obj in objs:
        if obj.is_output:
            count += 1
        elif bt.is_storage(obj):
            count += 1
    return count


def count_input_storages(objs):
    count = 0
    for obj in objs:
        if bt.is_input_storage(obj):
            count += 1
    return count


def count_output_storages(objs):
    count = 0
    for obj in objs:
        if obj.is_output:
            count += 1
    return count


def set_default_dtype(config, objs):
    default_datawidth = config['default_datawidth']
    for obj in objs:
        if obj.dtype is None:
            obj.dtype = dtype_list.dtype_info('int', default_datawidth)


def sence_edge(m, clk, wire, rst=None, mode='posedge', name='sence'):
    tmp_reg = m.TmpRegLike(wire, prefix=name, initval=0)
    edge_seq = vg.TmpSeq(m, clk, rst, prefix='sence_edge')
    edge_seq(tmp_reg(wire))

    edge = m.TmpWireLike(wire, prefix=name)
    if mode == 'posedge':
        edge.assign(vg.And(vg.Not(tmp_reg), wire))
    elif mode == 'negedge':
        edge.assign(vg.And(tmp_reg, vg.Not(wire)))
    else:
        raise ValueError(mode)

    return edge


def add_irq(m, clk, rst, saxi, bit_field, target_signal, mode='posedge', name=None):
    irq_target = sence_edge(m, clk, target_signal, rst, mode=mode, name=name)
    saxi.seq.If(irq_target)(
        saxi.register[control_reg_interrupt_isr][bit_field](irq_target)
    )


def make_module(config, name, objs, num_storages, num_input_storages, num_output_storages):
    m = vg.Module(name)

    clk = m.Input(config['clock_name'])

    if 'low' in config['reset_polarity'].lower():
        rst_seq = vg.Seq(m, 'rst_seq', clk)
        rst_x = m.Input(config['reset_name'])
        rst_x_inv = m.Wire(config['reset_name'] + '_inv')
        rst_x_inv.assign(vg.Not(rst_x))
        rst = m.Wire(config['reset_name'] + '_inv_buf')
        rst.assign(rst_seq.Prev(rst_x_inv, 2))

    else:
        rst = m.Input(config['reset_name'])

    if config['interrupt_enable']:
        irq = m.OutputReg(config['interrupt_name'], initval=0)

    datawidth = config['maxi_datawidth']
    addrwidth = config['maxi_addrwidth']
    cache_mode = axi.AxCACHE_COHERENT if config['axi_coherent'] else axi.AxCACHE_NONCOHERENT
    prot_mode = axi.AxPROT_COHERENT if config['axi_coherent'] else axi.AxPROT_NONCOHERENT
    user_mode = axi.AxUSER_COHERENT if config['axi_coherent'] else axi.AxUSER_NONCOHERENT

    maxi = vthread.AXIM(m, 'maxi', clk, rst, datawidth, addrwidth,
                        waddr_cache_mode=cache_mode, raddr_cache_mode=cache_mode,
                        waddr_prot_mode=prot_mode, raddr_prot_mode=prot_mode,
                        waddr_user_mode=user_mode, raddr_user_mode=user_mode,
                        use_global_base_addr=True,
                        req_fifo_addrwidth=config['req_fifo_addrwidth'],
                        fsm_as_module=config['fsm_as_module'])

    datawidth = config['saxi_datawidth']
    addrwidth = config['saxi_addrwidth']

    num_unified_storages = 1
    num_temporal_storages = 1

    if config['use_map_ram']:
        length = num_header_regs + num_control_regs + num_addr_map_regs
    elif not config['use_map_reg']:
        length = (num_header_regs + num_control_regs +
                  num_temporal_storages + num_input_storages + num_output_storages +
                  num_unified_storages)
    else:
        length = (num_header_regs + num_control_regs +
                  num_temporal_storages + num_storages)

    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst,
                                    datawidth, addrwidth, length=length,
                                    fsm_as_module=config['fsm_as_module'])

    maxi_idle = m.Wire('maxi_idle')
    maxi_idle.assign(vg.And(maxi.write_idle, maxi.read_idle))

    sw_rst_logic = m.Wire('sw_rst_logic')
    sw_rst_logic.assign(vg.And(maxi_idle, saxi.register[control_reg_reset]))

    rst_logic = m.Wire('rst_logic')
    rst_logic.assign(vg.Or(rst, sw_rst_logic))

    sys_rst_seq = vg.Seq(m, 'sys_rst_seq', clk)
    sys_rst = m.Reg('RST')
    v = rst_logic
    for i in range(2):
        v = vg.Or(v, sys_rst_seq.Prev(rst_logic, i + 1))
    sys_rst_seq(
        sys_rst(v)
    )

    if config['interrupt_enable']:
        irq_tmp = m.TmpWireLike(saxi.register[control_reg_interrupt_isr], prefix="irq")
        irq_tmp.assign(vg.And(saxi.register[control_reg_interrupt_isr],
                              saxi.register[control_reg_interrupt_ier]))
        irq_seq = vg.Seq(m, 'interrupt_seq', clk, rst)
        irq_seq(
            irq(vg.Uor(irq_tmp))
        )

        list(map(lambda ack, stat: saxi.seq.If(ack == 1)(ack(0), stat(0)),
                 saxi.register[control_reg_interrupt_iar], saxi.register[control_reg_interrupt_isr]))

        irq_busy = m.Wire("irq_busy")
        irq_busy.assign(saxi.register[control_reg_busy][0])
        add_irq(m, clk, rst, saxi, control_reg_interrupt_isr_busy,
                irq_busy, mode='negedge', name="irq_busy_edge")

        irq_extern = m.Wire("irq_extern")
        irq_extern.assign(vg.Uor(saxi.register[control_reg_extern_send]))
        add_irq(m, clk, rst, saxi, control_reg_interrupt_isr_extern,
                irq_extern, name="irq_extern_edge")

    for obj in objs:
        obj.set_module_info(m, clk, sys_rst, maxi, saxi)

    return m, clk, sys_rst, maxi, saxi


def schedule(config, objs):
    s = scheduler.OperationScheduler(config)
    s.schedule(objs)
    return s.result


def make_header_addr_map(config, saxi):
    header_info = collections.OrderedDict()
    header_regs = saxi.register[:num_header_regs]

    for i in range(num_header_regs):
        name = 'header%d' % i
        initval = config[name] if name in config else 0
        header_regs[i].initval = initval
        header_info[i] = 'header{} (default: 0x{:08x})'.format(i, initval)

    return header_info


def allocate(config, m, clk, rst, maxi, saxi, objs, schedule_table):
    set_storage_name(objs)
    merge_shared_attrs(objs)

    max_stream_rams = calc_max_stream_rams(config, schedule_table)
    ram_dict = make_rams(config, m, clk, rst, maxi, schedule_table, max_stream_rams)
    ram_set_cache = make_ram_sets(config, schedule_table, ram_dict, max_stream_rams)

    control_param_dict = make_control_params(config, schedule_table)

    substrm_dict = make_substreams(config, m, clk, rst, maxi, schedule_table)
    stream_cache = make_streams(config, schedule_table, ram_dict, substrm_dict)

    (global_addr_map, local_addr_map,
     global_map_info, global_mem_map) = make_addr_map(config, objs, saxi)

    if config['use_map_ram']:
        global_map_ram, local_map_ram = make_addr_map_rams(config, m, clk, rst, maxi,
                                                           global_addr_map, local_addr_map)
    else:
        global_map_ram = None
        local_map_ram = None

    control_cache, main_fsm = make_controls(
        config, m, clk, rst, maxi, saxi,
        schedule_table, control_param_dict,
        global_addr_map, local_addr_map,
        global_map_ram, local_map_ram)

    return (ram_dict, substrm_dict, ram_set_cache, stream_cache, control_cache,
            main_fsm, global_map_info, global_mem_map)


def set_storage_name(objs):
    tmp_input = 0
    tmp_output = 0
    for obj in objs:
        if bt.is_storage(obj) and obj.name is None:
            obj.name = 'input_%d' % tmp_input
            tmp_input += 1
        elif obj.is_output and obj.name is None:
            obj.name = 'output_%s_%d' % (obj.__class__.__name__, tmp_output)
            tmp_output += 1


def merge_shared_attrs(objs):
    repr_obj_dict = collections.OrderedDict()

    for obj in objs:
        if not bt.is_operator(obj):
            continue

        stream_hash = obj.get_stream_hash()
        if stream_hash not in repr_obj_dict:
            obj.merge_shared_attrs(obj)
            repr_obj_dict[stream_hash] = obj
            continue

        orig = repr_obj_dict[stream_hash]
        orig.merge_shared_attrs(obj)


def calc_max_stream_rams(config, schedule_table):
    max_stream_rams = {}  # key: stream_hash, value: max RAM sizes

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):

        for obj in objs:
            if not bt.is_operator(obj):
                continue

            if (bt.is_output_chainable_operator(obj) and
                    not obj.chain_head):
                continue

            input_rams, output_rams, temp_rams = obj.get_required_rams()
            stream_hash = obj.get_stream_hash()

            if stream_hash not in max_stream_rams:
                max_stream_rams[stream_hash] = (input_rams, output_rams, temp_rams)
            else:
                prev_input_rams, prev_output_rams, prev_temp_rams = max_stream_rams[stream_hash]
                new_input_rams = tuple([max_tuple(prev, cur)
                                        for prev, cur in zip(prev_input_rams, input_rams)])
                new_output_rams = tuple([max_tuple(prev, cur)
                                         for prev, cur in zip(prev_output_rams, output_rams)])
                new_temp_rams = tuple([max_tuple(prev, cur)
                                       for prev, cur in zip(prev_temp_rams, temp_rams)])
                max_stream_rams[stream_hash] = (new_input_rams, new_output_rams, new_temp_rams)

    return max_stream_rams


def make_rams(config, m, clk, rst, maxi, schedule_table, max_stream_rams):
    max_rams = calc_max_rams(config, schedule_table, max_stream_rams)

    ram_dict = collections.defaultdict(list)
    ram_index = collections.defaultdict(int)

    for (width, length), num in reversed(sorted(max_rams.items(),
                                                key=lambda x: x[0][0] * x[0][1])):
        for _ in range(num):
            datawidth = width
            numports = 2
            numbanks = maxi.datawidth // width

            key = (width, length)
            i = ram_index[key]
            ram_index[key] += 1

            name = 'ram_w%d_l%d_id%d' % (width, length, i)

            if numbanks <= 1:
                addrwidth = int(math.ceil(math.log(length, 2)))
                ram = vthread.RAM(m, name, clk, rst,
                                  datawidth, addrwidth, numports=numports,
                                  ram_style=config['onchip_ram_style'])
            else:
                addrwidth = int(math.ceil(math.log(length // numbanks, 2)))
                ram = vthread.MultibankRAM(m, name, clk, rst,
                                           datawidth, addrwidth, numports=numports,
                                           numbanks=numbanks,
                                           ram_style=config['onchip_ram_style'])

            ram_dict[key].append(ram)

    return ram_dict


def calc_max_rams(config, schedule_table, max_stream_rams):
    max_stage_rams = calc_max_stage_rams(config, schedule_table, max_stream_rams)

    priority = collections.Counter()

    if config['onchip_ram_priority'] == 'total_size':
        method = (lambda cur, width, length, num:
                  cur + width * length * num)
    elif config['onchip_ram_priority'] == 'max_size':
        method = (lambda cur, width, length, num:
                  max(cur, width * length))
    elif callable(config['onchip_ram_priority']):
        method = config['onchip_ram_priority']
    else:
        raise ValueError("Unsupported onchip_ram_priority '%s'" %
                         str(config['onchip_ram_priority']))

    for stage, rams in sorted(max_stage_rams.items(), key=lambda x: x[0]):
        for (width, length), num in rams.items():
            priority[stage] = method(priority[stage], width, length, num)

    max_rams = collections.Counter()

    for stage, rams in sorted(max_stage_rams.items(),
                              key=lambda x: priority[x[0]],
                              reverse=True):

        used_count = collections.Counter()

        for (width, length), num in sorted(rams.items(),
                                           key=lambda x: x[0][0] * x[0][1],
                                           reverse=True):

            for (mw, ml), mn in sorted(max_rams.items(),
                                       key=lambda x: x[0][0] * x[0][1],
                                       reverse=True):
                if mw != width:
                    continue

                if ml >= length and used_count[(mw, ml)] + num <= mn:
                    used_count[(mw, ml)] += num
                    break

            else:
                max_rams[(width, length)] += num
                used_count[(width, length)] += num

    return max_rams


def calc_max_stage_rams(config, schedule_table, max_stream_rams):
    max_stage_rams = {}  # key: stage_number, value: RAM counter

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):
        cnt = collections.Counter()

        for obj in objs:
            if not bt.is_operator(obj):
                continue

            if (bt.is_output_chainable_operator(obj) and
                    not obj.chain_head):
                continue

            stream_hash = obj.get_stream_hash()
            input_rams, output_rams, temp_rams = max_stream_rams[stream_hash]

            for width, length in sorted(input_rams, reverse=True):
                key = to_actual_ram_spec(config, width, length)
                cnt[key] += 1

            for width, length in sorted(output_rams, reverse=True):
                key = to_actual_ram_spec(config, width, length)
                cnt[key] += 1

            for width, length in sorted(temp_rams, reverse=True):
                key = to_actual_ram_spec(config, width, length)
                cnt[key] += 1

        max_stage_rams[stage] = cnt

    return max_stage_rams


def max_tuple(*tuples):
    return tuple([max(*values) for values in zip(*tuples)])


def make_ram_sets(config, schedule_table, ram_dict, max_stream_rams):
    ram_set_cache = collections.defaultdict(list)

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):
        make_stage_ram_sets(config, schedule_table, ram_dict, max_stream_rams,
                            stage, objs, ram_set_cache)

    return ram_set_cache


def make_stage_ram_sets(config, schedule_table, ram_dict, max_stream_rams,
                        stage, objs, ram_set_cache):

    ram_index_set = collections.defaultdict(set)

    # cache check
    for obj in objs:
        if not bt.is_operator(obj):
            continue

        if bt.is_output_chainable_operator(obj) and not obj.chain_head:
            continue

        stream_hash = obj.get_stream_hash()

        for (input_rams, output_rams,
             temp_rams, used_ram_index_dict) in ram_set_cache[stream_hash]:

            satisfied = obj.check_ram_requirements(input_rams,
                                                   output_rams, temp_rams)

            for key, ram_indexes in used_ram_index_dict.items():
                for ram_index in ram_indexes:
                    if ram_index in ram_index_set[key]:
                        satisfied = False

            # Hit: reuse the existing RAM set
            if satisfied:
                obj.set_rams(input_rams, output_rams, temp_rams)
                obj.cached_ram_set = True

                for key, ram_indexes in used_ram_index_dict.items():
                    for ram_index in ram_indexes:
                        ram_index_set[key].add(ram_index)

    # Miss or Unsatisfied: create a new RAM set
    for obj in objs:
        if not bt.is_operator(obj):
            continue

        if bt.is_output_chainable_operator(obj) and not obj.chain_head:
            continue

        if obj.cached_ram_set:
            continue

        stream_hash = obj.get_stream_hash()
        req_inputs, req_outputs, req_temps = max_stream_rams[stream_hash]
        used_ram_index_dict = collections.defaultdict(list)

        req_rams = []
        req_rams.extend([(width, length, 'input', i)
                         for i, (width, length) in enumerate(req_inputs)])
        req_rams.extend([(width, length, 'output', i)
                         for i, (width, length) in enumerate(req_outputs)])
        req_rams.extend([(width, length, 'temp', i)
                         for i, (width, length) in enumerate(req_temps)])

        input_rams = [None for _ in req_inputs]
        output_rams = [None for _ in req_outputs]
        temp_rams = [None for _ in req_temps]

        # smallest request first
        for width, length, ram_type, pos in sorted(req_rams,
                                                   key=lambda x: (x[0], x[1])):
            width, length = to_actual_ram_spec(config, width, length)
            key = (width, length)

            found = False
            # smallest RAM first
            for ram_key, rams in sorted(ram_dict.items(), key=lambda x: x[0]):
                if found:
                    break

                for i, ram in enumerate(rams):
                    if i in ram_index_set[ram_key]:
                        continue

                    ram_width, ram_length = ram_key

                    if ram_width != width:
                        continue

                    if ram_length >= length:
                        if ram_type == 'input':
                            input_rams[pos] = ram
                        elif ram_type == 'output':
                            output_rams[pos] = ram
                        elif ram_type == 'temp':
                            temp_rams[pos] = ram

                        ram_index_set[ram_key].add(i)
                        used_ram_index_dict[key].append(i)
                        found = True
                        break

        obj.set_rams(input_rams, output_rams, temp_rams)
        ram_set_cache[stream_hash].append((input_rams, output_rams,
                                           temp_rams, used_ram_index_dict))


def to_actual_ram_spec(config, width, length):
    if width == 0:
        width = config['default_datawidth']

    min_capacity = config['min_onchip_ram_capacity']
    maxi_datawidth = config['maxi_datawidth']
    numbanks = maxi_datawidth // width

    if numbanks < 1:
        numbanks = 1

    # while min_length is the total length of a MultibankMemory,
    # min_capacity is the minimum capacity for each bank
    min_length = (min_capacity // width) * numbanks

    if length < min_length:
        length = min_length

    addrwidth = int(math.ceil(math.log(length, 2)))
    real_length = 2 ** addrwidth

    return width, real_length


def make_control_params(config, schedule_table):

    control_param_dict = get_control_param_dict(schedule_table)
    obj_cache = collections.defaultdict(list)

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):

        for obj in objs:
            if not bt.is_operator(obj):
                continue

            if bt.is_view(obj):
                continue

            if bt.is_removable_reshape(obj):
                continue

            if (bt.is_output_chainable_operator(obj) and
                    not obj.chain_head):
                continue

            key = obj.get_stream_hash()
            control_param_list = control_param_dict[key]
            control_param_len = len(control_param_list)

            if (not config['disable_control_cache'] and
                    obj.control_cachable and len(obj_cache[key]) > 0):
                # hit
                orig = obj_cache[key][0]
                obj.copy_control_params(orig)

            else:
                # miss
                width_dict, signed_dict = calc_control_param_width(control_param_list)
                obj.make_control_params(control_param_len, width_dict, signed_dict,
                                        use_param_ram=config['use_param_ram'],
                                        min_param_ram_len=config['min_param_ram_len'])
                obj.make_control_param_buf(control_param_list,
                                           use_param_ram=config['use_param_ram'],
                                           min_param_ram_len=config['min_param_ram_len'],
                                           ram_style=config['param_ram_style'])

                obj_cache[key].append(obj)

    return control_param_dict


def get_control_param_dict(schedule_table):
    index_dict = collections.defaultdict(int)
    control_param_dict = collections.defaultdict(list)

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):

        for obj in objs:
            if not bt.is_operator(obj):
                continue

            if bt.is_view(obj):
                continue

            if bt.is_removable_reshape(obj):
                continue

            if (bt.is_output_chainable_operator(obj) and
                    not obj.chain_head):
                continue

            key = obj.get_stream_hash()
            index = index_dict[key]
            obj.set_control_param_index(index)
            index_dict[key] += 1

            values = obj.collect_all_control_param_values()
            control_param_dict[key].append(values)

    return control_param_dict


def calc_control_param_width(control_param_value_list):
    width_dict = collections.OrderedDict()
    signed_dict = collections.OrderedDict()

    for values in control_param_value_list:
        for name, value in values.items():
            if isinstance(value, (tuple, list)):
                if name in width_dict:
                    current_width = width_dict[name]
                else:
                    current_width = [1 for _ in value]
                if name in signed_dict:
                    current_signed = signed_dict[name]
                else:
                    current_signed = [False for _ in value]

                width = []
                signed = []
                for v, cw, cs in zip(value, current_width, current_signed):
                    w = max(abs(v).bit_length(), 1)
                    width.append(max(cw, w))
                    s = v < 0
                    signed.append(cs or s)

                width_dict[name] = width
                signed_dict[name] = signed

            else:
                if name in width_dict:
                    current_width = width_dict[name]
                else:
                    current_width = 1
                if name in signed_dict:
                    current_signed = signed_dict[name]
                else:
                    current_signed = False

                w = max(abs(value).bit_length(), 1)
                width = max(current_width, w)
                s = value < 0
                signed = current_signed or s

                width_dict[name] = width
                signed_dict[name] = signed

    for name, signed in signed_dict.items():
        if isinstance(width_dict[name], (tuple, list)):
            width = []
            for w, s in zip(width_dict[name], signed):
                if s:
                    width.append(w + 1)
                else:
                    width.append(w)
            width_dict[name] = width
        else:
            if signed:
                width_dict[name] += 1

    return width_dict, signed_dict


def make_substreams(config, m, clk, rst, maxi, schedule_table):
    max_substrms = calc_max_substreams(config, schedule_table)

    substrm_dict = collections.defaultdict(list)
    substrm_index = collections.defaultdict(int)

    for key, num in sorted(max_substrms.items(), key=lambda x: x[0]):
        method_name = key[0]
        args = key[1]
        method = getattr(substreams, method_name)

        for _ in range(num):
            i = substrm_index[key]
            substrm_index[key] += 1
            substrm = method(m, clk, rst, *args)
            substrm_dict[key].append(substrm)

    return substrm_dict


def calc_max_substreams(config, schedule_table):
    max_substrms = collections.Counter()

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):
        cnt = collections.Counter()

        for obj in objs:
            if not bt.is_operator(obj):
                continue

            if (bt.is_output_chainable_operator(obj) and
                    not obj.chain_head):
                continue

            substrms = obj.get_required_substreams()

            for key in substrms:
                cnt[key] += 1

        for key, val in sorted(cnt.items(), key=lambda x: x[0], reverse=True):
            max_substrms[key] = max(max_substrms[key], val)

    return max_substrms


def to_hashable_dict(dct):
    return tuple([(key, val)
                  for key, val in sorted(dct.items(), key=lambda x:x[0])])


def from_hashable_dict(dct):
    return dict(dct)


def make_streams(config, schedule_table, ram_dict, substrm_dict):
    stream_cache = collections.defaultdict(list)

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):
        make_stage_streams(config, schedule_table, ram_dict, substrm_dict,
                           stage, objs, stream_cache)

    return stream_cache


def make_stage_streams(config, schedule_table, ram_dict, substrm_dict,
                       stage, objs, stream_cache):

    substrm_index_set = collections.defaultdict(set)

    # cache check
    for obj in objs:
        if config['disable_stream_cache']:
            break

        if not bt.is_operator(obj):
            continue

        if bt.is_output_chainable_operator(obj) and not obj.chain_head:
            continue

        if not obj.stream_cachable:
            continue

        stream_hash = obj.get_stream_hash()

        for (strm, used_substrm_index_dict) in stream_cache[stream_hash]:

            satisfied = True

            for key, substrm_indexes in used_substrm_index_dict.items():
                for substrm_index in substrm_indexes:
                    if substrm_index in substrm_index_set:
                        satisfied = False

            # Hit: reuse the existing substream set and main stream
            if satisfied:
                obj.set_stream(strm)
                obj.cached_stream = True

                for key, substrm_indexes in used_substrm_index_dict.items():
                    for substrm_index in substrm_indexes:
                        substrm_index_set[key].add(substrm_index)

    # Miss: create a new substream set and main stream
    for obj in objs:
        if not bt.is_operator(obj):
            continue

        if bt.is_output_chainable_operator(obj) and not obj.chain_head:
            continue

        if obj.cached_stream:
            continue

        req_substrms = obj.get_required_substreams()
        used_substrm_index_dict = collections.defaultdict(list)

        substrms = []
        for key in req_substrms:

            substrm_list = substrm_dict[key]
            for i, substrm in enumerate(substrm_list):
                if i in substrm_index_set[key]:
                    continue

                substrm_index_set[key].add(i)
                used_substrm_index_dict[key].append(i)
                sub = substrm_list[i]
                substrms.append(sub)
                break

        obj.set_substreams(substrms)
        strm = obj.make_stream(datawidth=config['default_datawidth'],
                               fsm_as_module=config['fsm_as_module'],
                               dump=config['dump_stream'],
                               dump_base=config['dump_stream_base'])
        obj.set_stream(strm)

        stream_hash = obj.get_stream_hash()
        stream_cache[stream_hash].append((strm, used_substrm_index_dict))


def make_addr_map(config, objs, saxi):

    chunk_size = config['offchipram_chunk_bytes']

    maxi_datawidth = config['maxi_datawidth']
    if (chunk_size * 8) % maxi_datawidth != 0:
        raise ValueError("'offchipram_chunk_bytes' must be a multiple number of 'maxi_datawidth'.")

    global_addr_offset = config['default_global_addr_offset']

    global_addr_map = collections.OrderedDict()
    local_addr_map = collections.OrderedDict()
    global_map_info = collections.OrderedDict()
    global_mem_map = collections.OrderedDict()  # key: (start, end)

    if not config['use_map_ram']:
        map_regs = saxi.register[num_header_regs + num_control_regs:]

    offset_reg = saxi.register[control_reg_global_offset]
    offset_reg.initval = global_addr_offset

    reg_index = 1
    global_index = 1
    local_index = 1

    storage_used = 0
    unified_storage_used = 0
    temporal_used = 0

    unified_storage_list = []

    local_addr_map[0] = 0

    # output
    for obj in sorted(objs, key=lambda x: x.object_id):
        if obj.is_output and obj.global_index is None:

            orig_obj = obj
            while bt.is_view(obj) or bt.is_removable_reshape(obj):
                obj = obj.args[0]

            width = obj.dtype.width
            length = obj.get_aligned_length()
            space_size = align_space(width, length, chunk_size)
            default_global_addr = storage_used

            obj.set_global_index(global_index)
            obj.set_local_index(0)
            obj.set_default_global_addr(default_global_addr)
            obj.set_default_local_addr(0)

            i = (("output (%s) %s "
                  "(size: %s, dtype: %s, shape: %s, "
                  "alignment: %d words (%d bytes)), "
                  "aligned shape: %s") %
                 (orig_obj.__class__.__name__,
                  "'%s'" % orig_obj.name if orig_obj.name is not None else 'None',
                  size_str(space_size),
                  orig_obj.dtype.to_str() if orig_obj.dtype is not None else 'None',
                  (str(orig_obj.shape)
                   if isinstance(orig_obj.shape, (tuple, list)) else '()'),
                  orig_obj.get_word_alignment(),
                  bt.to_byte(orig_obj.get_word_alignment() * orig_obj.get_ram_width()),
                  (str(tuple(orig_obj.get_aligned_shape()))
                   if isinstance(orig_obj.shape, (tuple, list)) else '()')))

            global_mem_map[(default_global_addr,
                            default_global_addr + space_size - 1)] = i

            if config['use_map_ram']:
                global_addr_map[global_index] = default_global_addr

            if not config['use_map_ram']:
                map_regs[global_index].initval = default_global_addr

            global_map_info[global_index] = i
            global_index += 1

            storage_used += space_size

    # input (placeholder)
    for obj in sorted(objs, key=lambda x: x.object_id):
        if not bt.is_operator(obj):
            continue

        if (bt.is_output_chainable_operator(obj) and
                not obj.chain_head):
            continue

        srcs = obj.collect_sources()

        for src in srcs:
            if src.global_index is not None:
                continue

            if bt.is_view(src):
                continue

            if bt.is_removable_reshape(src):
                continue

            if not bt.is_input_storage(src):
                continue

            # source
            width = src.dtype.width
            length = src.get_aligned_length()
            space_size = align_space(width, length, chunk_size)
            default_global_addr = storage_used

            src.set_global_index(global_index)
            src.set_local_index(0)
            src.set_default_global_addr(default_global_addr)
            src.set_default_local_addr(0)

            i = (("%s %s "
                  "(size: %s, dtype: %s, shape: %s, "
                  "alignment: %d words (%d bytes)), "
                  "aligned shape: %s") %
                 (src.__class__.__name__,
                  "'%s'" % src.name if src.name is not None else 'None',
                  size_str(space_size),
                  src.dtype.to_str() if src.dtype is not None else 'None',
                  (str(src.shape)
                   if isinstance(src.shape, (tuple, list)) else '()'),
                  src.get_word_alignment(),
                  bt.to_byte(src.get_word_alignment()
                             * src.get_ram_width()),
                  (str(tuple(src.get_aligned_shape()))
                   if isinstance(src.shape, (tuple, list)) else '()')))

            global_mem_map[(default_global_addr,
                            default_global_addr + space_size - 1)] = i

            if config['use_map_ram']:
                global_addr_map[global_index] = default_global_addr

            if not config['use_map_ram']:
                map_regs[global_index].initval = default_global_addr

            global_map_info[global_index] = i
            global_index += 1

            storage_used += space_size

    # unified input (variable, constant)
    if not config['use_map_ram'] and not config['use_map_reg']:
        unified_global_index = global_index
        unified_default_global_addr = storage_used
        map_regs[unified_global_index].initval = unified_default_global_addr
        global_index += 1

    for obj in sorted(objs, key=lambda x: x.object_id):
        if not bt.is_operator(obj):
            continue

        if (bt.is_output_chainable_operator(obj) and
                not obj.chain_head):
            continue

        srcs = obj.collect_sources()

        for src in srcs:
            if src.global_index is not None:
                continue

            if bt.is_view(src):
                continue

            if bt.is_removable_reshape(src):
                continue

            if not bt.is_storage(src):
                continue

            # source
            width = src.dtype.width
            length = src.get_aligned_length()
            space_size = align_space(width, length, chunk_size)
            default_global_addr = storage_used

            if not config['use_map_ram'] and not config['use_map_reg']:
                src.set_global_index(unified_global_index)
                src.set_local_index(0)
                src.set_default_global_addr(unified_default_global_addr)
                src.set_default_local_addr(unified_storage_used)
            else:
                src.set_global_index(global_index)
                src.set_local_index(0)
                src.set_default_global_addr(default_global_addr)
                src.set_default_local_addr(0)

            i = (("%s %s "
                  "(size: %s, dtype: %s, shape: %s, "
                  "alignment: %d words (%d bytes)), "
                  "aligned shape: %s") %
                 (src.__class__.__name__,
                  "'%s'" % src.name if src.name is not None else 'None',
                  size_str(space_size),
                  src.dtype.to_str() if src.dtype is not None else 'None',
                  (str(src.shape)
                   if isinstance(src.shape, (tuple, list)) else '()'),
                  src.get_word_alignment(),
                  bt.to_byte(src.get_word_alignment()
                             * src.get_ram_width()),
                  (str(tuple(src.get_aligned_shape()))
                   if isinstance(src.shape, (tuple, list)) else '()')))

            global_mem_map[(default_global_addr,
                            default_global_addr + space_size - 1)] = i

            if config['use_map_ram']:
                global_addr_map[global_index] = default_global_addr

            if not config['use_map_ram'] and config['use_map_reg']:
                map_regs[global_index].initval = default_global_addr

            if config['use_map_ram']:
                global_map_info[global_index] = i

            global_index += 1

            storage_used += space_size
            unified_storage_used += space_size
            unified_storage_list.append(src)

    if (not config['use_map_ram'] and not config['use_map_reg'] and
            unified_storage_used > 0):

        names = ', '.join(["'%s'" % obj.name if obj.name is not None else 'None'
                           for obj in unified_storage_list])
        i = ('variables %s (size: %s)' %
             (names, size_str(unified_storage_used)))
        global_map_info[unified_global_index] = i

    # temporal
    for obj in sorted(objs, key=lambda x: x.object_id):
        if not bt.is_operator(obj):
            continue

        if (bt.is_output_chainable_operator(obj) and
                not obj.chain_head):
            continue

        srcs = obj.collect_sources()

        for src in srcs:
            if src.global_index is not None:
                continue

            if bt.is_view(src):
                continue

            if bt.is_removable_reshape(src):
                continue

            if bt.is_storage(src):
                continue

            # temporal
            width = src.dtype.width
            length = src.get_aligned_length()
            space_size = align_space(width, length, chunk_size)

            default_global_addr = storage_used
            default_local_addr = temporal_used
            local_addr_map[local_index] = default_local_addr

            src.set_global_index(0)
            src.set_local_index(local_index)
            src.set_default_global_addr(default_global_addr)
            src.set_default_local_addr(default_local_addr)

            local_index += 1
            temporal_used += space_size

    default_global_addr = storage_used
    global_addr_map[0] = default_global_addr

    i = ('temporal storages (size: %s)' %
         size_str(temporal_used))

    global_mem_map[(default_global_addr,
                    default_global_addr + temporal_used - 1)] = i

    if not config['use_map_ram']:
        map_regs[0].initval = default_global_addr

    global_map_info[0] = i

    return global_addr_map, local_addr_map, global_map_info, global_mem_map


def align_space(width, length, chunk_size):
    bytes = int(math.ceil((width * length) / 8))
    num_chunks = int(math.ceil(bytes / chunk_size))
    return chunk_size * num_chunks


def size_str(num_bytes):
    if num_bytes >= 1024:
        return '%dKB' % int(math.ceil(num_bytes / 1024))

    return '%dB' % num_bytes


def make_addr_map_rams(config, m, clk, rst, maxi,
                       global_addr_map, local_addr_map):

    datawidth = config['maxi_addrwidth']

    global_name = 'global_map_ram'

    if len(global_addr_map) > 1:
        global_addrwidth = int(math.ceil(math.log(len(global_addr_map), 2)))
    else:
        global_addrwidth = 1

    global_initvals = []
    for index, addr in sorted(global_addr_map.items(), key=lambda x: x[0]):
        global_initvals.append(addr)

    global_map_ram = vthread.RAM(m, global_name, clk, rst,
                                 datawidth, global_addrwidth, numports=2,
                                 initvals=global_initvals,
                                 ram_style=config['map_ram_style'])

    local_name = 'local_map_ram'

    if len(local_addr_map) > 1:
        local_addrwidth = int(math.ceil(math.log(len(local_addr_map), 2)))
    else:
        local_addrwidth = 1

    local_initvals = []
    for index, addr in sorted(local_addr_map.items(), key=lambda x: x[0]):
        local_initvals.append(addr)

    local_map_ram = vthread.RAM(m, local_name, clk, rst,
                                datawidth, local_addrwidth, numports=1,
                                initvals=local_initvals,
                                ram_style=config['map_ram_style'])

    return global_map_ram, local_map_ram


def make_controls(config, m, clk, rst, maxi, saxi,
                  schedule_table, control_param_dict,
                  global_addr_map, local_addr_map,
                  global_map_ram, local_map_ram):

    num_global_vars = len(global_addr_map)

    name = 'main_fsm'
    main_fsm = vg.FSM(m, name, clk, rst, as_module=config['fsm_as_module'])
    state_init = main_fsm.current
    main_fsm.state_list = []

    saxi.seq.If(main_fsm.state == state_init)(
        saxi.register[control_reg_busy](0),
        saxi.register[control_reg_reset](0),
        saxi.register[control_reg_extern_send](0)
    )

    if config['measurable_main_fsm']:
        internal_counter = m.Reg("internal_state_counter", width=32, initval=0)
        saxi.seq.If(main_fsm.state == state_init + 1)(
            internal_counter(0),
            saxi.register[control_reg_count](0)
        ).Elif(main_fsm.state == saxi.register[control_reg_count_state])(
            vg.If(internal_counter == saxi.register[control_reg_count_div])(
                internal_counter(0),
                saxi.register[control_reg_count](saxi.register[control_reg_count] + 1)
            ).Else(
                internal_counter(internal_counter + 1)
            )
        )

    if not config['use_map_ram']:
        map_regs = saxi.register[num_header_regs + num_control_regs:]

    offset_reg = saxi.register[control_reg_global_offset]
    maxi.seq(
        maxi.global_base_addr(offset_reg)
    )

    if config['use_map_ram']:
        start = saxi.register[control_reg_start]
        load_global_addr_map = saxi.register[control_reg_load_global_addr_map]
        global_addr = saxi.register[control_reg_addr_global_addr_map]

        def wait_start():
            while True:
                if load_global_addr_map:
                    saxi.write(control_reg_busy_global_addr_map, 1)
                    saxi.write(control_reg_load_global_addr_map, 0)

                    maxi.dma_read(global_map_ram, 0, global_addr,
                                  num_global_vars, port=1)

                    saxi.write(control_reg_busy_global_addr_map, 0)

                elif start:
                    saxi.write(control_reg_busy, 1)
                    saxi.write(control_reg_start, 0)
                    break

        main_fsm = vthread.embed_thread(main_fsm, wait_start)

    else:
        saxi.wait(main_fsm, control_reg_start, 0, polarity=False)
        saxi.write(main_fsm, control_reg_busy, 1)
        saxi.write(main_fsm, control_reg_start, 0)

    control_cache = collections.defaultdict(list)

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):

        for obj in objs:
            if not bt.is_operator(obj):
                continue

            if bt.is_view(obj):
                continue

            if bt.is_removable_reshape(obj):
                continue

            if (bt.is_output_chainable_operator(obj) and
                    not obj.chain_head):
                continue

            param_key = obj.get_stream_hash()
            control_param_list = control_param_dict[param_key]
            control_param_len = len(control_param_list)

            key = obj.get_control_hash()

            if (not config['disable_control_cache'] and
                    obj.control_cachable and len(control_cache[key]) > 0):
                # hit
                control, orig = control_cache[key][0]
                if control.stream_ram_hash != key:
                    raise ValueError("hash mismatch: '%x' != '%x'" %
                                     (control.stream_ram_hash, key))

                obj.copy_control(orig)

            else:
                # miss
                obj.make_objaddr()
                obj.make_arg_objaddrs()

                control = obj.make_control(fsm_as_module=config['fsm_as_module'])
                control.stream_ram_hash = key
                control_cache[key].append((control, obj))

            # bind address
            if config['use_map_ram']:
                gaddr = global_map_ram.read(main_fsm, obj.global_index)
                main_fsm.set_index(main_fsm.current - 1)
                laddr = local_map_ram.read(main_fsm, obj.local_index)
            else:
                gaddr = map_regs[obj.global_index]
                laddr = obj.default_local_addr

            if laddr != 0:
                addr = gaddr + laddr
            else:
                addr = gaddr

            main_fsm(
                obj.objaddr(addr)
            )
            main_fsm.goto_next()

            arg_global_indexes = obj.get_arg_global_indexes()
            arg_local_indexes = obj.get_arg_local_indexes()
            arg_default_global_addrs = obj.get_arg_default_global_addrs()
            arg_default_local_addrs = obj.get_arg_default_local_addrs()

            for (arg_objaddr, arg_global_index, arg_local_index,
                 arg_default_global_addr, arg_default_local_addr) in zip(
                     obj.arg_objaddrs, arg_global_indexes, arg_local_indexes,
                     arg_default_global_addrs, arg_default_local_addrs):

                if config['use_map_ram']:
                    gaddr = global_map_ram.read(main_fsm, arg_global_index)
                    main_fsm.set_index(main_fsm.current - 1)
                    laddr = local_map_ram.read(main_fsm, arg_local_index)
                else:
                    gaddr = map_regs[arg_global_index]
                    laddr = arg_default_local_addr

                if laddr != 0:
                    addr = gaddr + laddr
                else:
                    addr = gaddr

                main_fsm(
                    arg_objaddr(addr)
                )
                main_fsm.goto_next()

            # bind parameter parameters
            obj.set_control_params(main_fsm, control_param_len,
                                   use_param_ram=config['use_param_ram'],
                                   min_param_ram_len=config['min_param_ram_len'])

            obj.run_control(main_fsm)

        ret_start_state = main_fsm.current
        main_fsm.goto_next()

        for obj in objs:
            if not bt.is_operator(obj):
                continue

            if (bt.is_output_chainable_operator(obj) and
                    not obj.chain_head):
                continue

            obj.join_control(main_fsm)
            obj.reset_control(main_fsm)

        ret_end_state = main_fsm.current

        if not bt.is_operator(obj):
            control_name = 'None'
        elif bt.is_view(obj):
            control_name = 'None'
        elif bt.is_removable_reshape(obj):
            control_name = 'None'
        elif (bt.is_output_chainable_operator(obj) and
                not obj.chain_head):
            control_name = 'None'
        else:
            control_name = control_cache[obj.get_control_hash()][0][0].name

        main_fsm.state_list.append(
            (ret_start_state, ret_end_state, obj.name, control_name))

        main_fsm.goto_next()

    # finalize
    main_fsm.goto_next()

    saxi.write(main_fsm, control_reg_busy, 0)

    main_fsm.goto_next()
    main_fsm.goto_next()
    main_fsm.goto_init()

    return control_cache, main_fsm


def make_reg_map(config, global_map_info, header_info):
    reg_map = collections.OrderedDict()
    reg_type = {'rw': 'RW', 'r': 'R ', 'w': ' W', 'x': '  '}

    for i in range(num_header_regs):
        index = index_to_bytes(i)
        reg_map[index] = (reg_type['r'], header_info[i])

    index = index_to_bytes(control_reg_start)
    reg_map[index] = (reg_type['w'], "Start (set '1' to run)")

    index = index_to_bytes(control_reg_busy)
    reg_map[index] = (reg_type['r'], "Busy (returns '1' when running)")

    index = index_to_bytes(control_reg_reset)
    reg_map[index] = (reg_type['w'], "Reset (set '1' to initialize internal logic)")

    index = index_to_bytes(control_reg_extern_send)
    reg_map[index] = (reg_type['r'], "Opcode from extern objects to SW (returns '0' when idle)")

    index = index_to_bytes(control_reg_extern_recv)
    reg_map[index] = (reg_type['w'], "Resume extern objects (set '1' to resume)")

    if config['interrupt_enable']:
        index = index_to_bytes(control_reg_interrupt_isr)
        reg_map[index] = (reg_type['r'], "Interrupt Status Register")

        index = index_to_bytes(control_reg_interrupt_ier)
        reg_map[index] = (reg_type['w'], "Interrupt Enable Register")

        index = index_to_bytes(control_reg_interrupt_iar)
        reg_map[index] = (reg_type['w'], "Interrupt Acknowledge Register")

    if config['use_map_ram']:
        control_reg_reserved_start = control_reg_addr_global_addr_map + 1
    elif config['interrupt_enable']:
        control_reg_reserved_start = control_reg_reserved
    else:
        control_reg_reserved_start = control_reg_interrupt_isr

    if config['measurable_main_fsm']:
        index = index_to_bytes(control_reg_count)
        reg_map[index] = (reg_type['r'], "State Counter")

        index = index_to_bytes(control_reg_count_state)
        reg_map[index] = (reg_type['w'], "Count Target")

        index = index_to_bytes(control_reg_count_div)
        reg_map[index] = (reg_type['w'], "Count Divider")
    else:
        control_reg_reserved_start = control_reg_count
    index = index_to_bytes(control_reg_reserved_start)
    reg_map[index] = (reg_type['x'], "Reserved ...")

    index = index_to_bytes(control_reg_address_amount - 1)
    reg_map[index] = (reg_type['x'], "... Reserved")

    index = index_to_bytes(control_reg_address_amount)
    reg_map[index] = (reg_type['r'], "Address space amount")

    index = index_to_bytes(control_reg_global_offset)
    default_global_addr_offset = config['default_global_addr_offset']
    reg_map[index] = (reg_type['rw'], 'Global address offset (default: %d)' %
                      default_global_addr_offset)

    if config['use_map_ram']:
        index = index_to_bytes(control_reg_load_global_addr_map)
        reg_map[index] = (reg_type['w'], "Load global address map (set '1' to load)")

        index = index_to_bytes(control_reg_busy_global_addr_map)
        reg_map[index] = (
            reg_type['r'], "Busy loading global address map (returns '1' when loading)")

        index = index_to_bytes(control_reg_addr_global_addr_map)
        reg_map[index] = (reg_type['w'], "Head address of global address map")

    else:
        for gindex, info in sorted(global_map_info.items(), key=lambda x: x[0]):
            index = index_to_bytes(control_reg_global_addr + gindex)
            reg_map[index] = (reg_type['rw'], 'Address of ' + info)

    return reg_map


def index_to_bytes(index, wordsize=4):
    return index * wordsize


def dump_main_fsm(main_fsm):
    s = []
    s.append('[State IDs in main_fsm]')

    for value in main_fsm.state_list:
        s.append("  %s" % str(value))

    print('\n'.join(s))


def dump_config(config, where_from=None, output=None):
    s = []
    s.append('NNgen: Neural Network Accelerator Generator (version %s)' %
             version.__version__)

    if where_from == 'to_verilog':
        s.append('[Verilog HDL]')
        s.append('  Output: %s' % output)
    elif where_from == 'to_ipxact':
        s.append('[IP-XACT]')
        s.append('  Output: %s' % output)

    s.append('[Configuration]')
    s.append('(AXI Master Interface)')
    s.append('  Data width   : %d' % config['maxi_datawidth'])
    s.append('  Address width: %d' % config['maxi_addrwidth'])
    s.append('(AXI Slave Interface)')
    s.append('  Data width   : %d' % config['saxi_datawidth'])
    s.append('  Address width: %d' % config['saxi_addrwidth'])

    no_print = ('maxi_datawidth', 'maxi_addrwidth',
                'saxi_datawidth', 'saxi_addrwidth')

    has_other_option = False
    for key, value in sorted(config.items(), key=lambda x: x[0]):
        if key in no_print:
            continue
        if key in default_config and default_config[key] != value:
            if not has_other_option:
                s.append('(Other Options)')
                has_other_option = True
            s.append('  %s: %s' % (key, str(value)))

    print('\n'.join(s))


def dump_schedule_table(schedule_table):
    s = []
    s.append('[Schedule Table]')

    for stage, objs in sorted(schedule_table.items(), key=lambda x: x[0]):
        s.append('(Stage %d)' % stage)
        for obj in objs:
            if bt.is_storage(obj):
                continue

            if bt.is_output_chainable_operator(obj) and not obj.chain_head:
                continue

            s.append('  %s' % str(obj))
            s.extend(dump_sources(obj, prefix='  | '))

    print('\n'.join(s))


def dump_sources(obj, prefix='  '):
    s = []
    srcs = obj.args
    for i, src in enumerate(srcs):
        s.append('%s%s' % (prefix, str(src)))
        if bt.is_output_chainable_operator(src) and not src.chain_head:
            next_prefix = prefix + '| '
            s.extend(dump_sources(src, prefix=next_prefix))
    return s


def dump_rams(ram_dict):
    s = []
    s.append('[RAM (spec: num)]')

    for (width, length), rams in sorted(ram_dict.items(), key=lambda x: x[0], reverse=True):
        num_rams = len(rams)
        if isinstance(rams[0], vthread.MultibankRAM):
            bank = rams[0].numbanks
            ports = rams[0].numports
        else:
            bank = 1
            ports = rams[0].numports

        s.append('  %d-bit %d-entry %d-port %d-bank RAM: %d' %
                 (width, length, ports, bank, num_rams))

    print('\n'.join(s))


def dump_substreams(substrm_dict):
    s = []
    s.append('[Substream (spec: num)]')

    for hsh, lst in substrm_dict.items():
        s.append('  %s: %d' % (str(hsh), len(lst)))

    print('\n'.join(s))


def dump_streams(stream_cache):
    s = []
    s.append('[Stream (spec: num)]')

    for hsh, lst in stream_cache.items():
        s.append('  %s: %d' % (str(hsh), len(lst)))

    print('\n'.join(s))


def dump_controls(control_cache, main_fsm):
    s = []
    s.append('[Control (name (# states: num))]')

    indexes = set(main_fsm.body.keys())
    indexes.update(set(main_fsm.jump.keys()))
    num_states = len(indexes)
    s.append('  %s (# states: %d)' % (main_fsm.name, num_states))

    for lst in control_cache.values():
        for control, obj in lst:
            indexes = set(control.fsm.body.keys())
            indexes.update(set(control.fsm.jump.keys()))
            num_states = len(indexes)
            s.append('  %s (# states: %d)' % (control.name, num_states))

    print('\n'.join(s))


def dump_register_map(reg_map):
    s = []
    s.append('[Register Map]')

    maximum = sorted(reg_map.items(), key=lambda x: x[0], reverse=True)[0][0]
    num_digits = max(int(math.ceil(math.log(maximum, 10))), 1)
    fmt = ''.join(('  %', '%d' % num_digits, 'd (%s): %s'))

    for i, (direction, desc) in sorted(reg_map.items(), key=lambda x: x[0]):
        s.append(fmt % (i, direction, desc))

    print('\n'.join(s))


def dump_memory_map(mem_map):
    max_gaddr = 0
    min_gaddr = 0
    for start, end in mem_map.keys():
        if start <= end:
            max_gaddr = max(max_gaddr, start, end)
            min_gaddr = min(min_gaddr, start, end)

    s = []
    num_bytes = max_gaddr - min_gaddr + 1
    s.append('[Default Memory Map (start - end)] (entire range: [%d - %d], size: %s)' %
             (min_gaddr, max_gaddr, size_str(num_bytes)))

    num_digits = max(int(math.ceil(math.log(max_gaddr, 10))), 1)
    fmt = ''.join(('  [%', '%d' % num_digits, 'd - %', '%d' % num_digits, 'd]: %s'))

    for (start, end), desc in sorted(mem_map.items(), key=lambda x: x[0]):
        if start <= end:
            s.append(fmt % (start, end, desc))

    print('\n'.join(s))



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.basic_types as bt
import nngen.storage as storage

from . import util
from . import conv2d
from . import matmul
from . import normalize
from . import sigmoid
from . import exp
from . import reduce


# describe custom quantize methods here
func_map = {
    'conv2d': conv2d.conv2d,
    'matmul': matmul.matmul,
    'normalize': normalize.normalize,
    'scaled_add': normalize.scaled_add,
    'scaled_concat': normalize.scaled_concat,
    'scaled_multiply': normalize.scaled_multiply,
    'scaled_div': normalize.scaled_div,
    'sigmoid': sigmoid.sigmoid,
    'exp': exp.exp,
    'argmax': reduce.argmax,
    'argmin': reduce.argmin,
}


def _has_func(op_type):
    return op_type in func_map


def _get_func(op_type):
    if op_type not in func_map:
        return None

    return func_map[op_type]


class _QuantizeVisitor(object):

    def __init__(self, input_scale_factors,
                 input_means=None, input_stds=None, input_generators=None, num_samples=1,
                 verbose=False):

        if not isinstance(input_scale_factors, dict):
            raise TypeError('input_scale_factors must be dict.')

        if input_means is None:
            input_means = {}

        if input_stds is None:
            input_stds = {}

        if input_generators is None:
            input_generators = {}

        self.input_scale_factors = input_scale_factors
        self.input_means = input_means
        self.input_stds = input_stds
        self.input_generators = input_generators
        self.num_samples = num_samples

        self.verbose = verbose

        self.memo = {}
        self.input_dict = {}

    def generic_visit(self, node):

        if not isinstance(node, bt._Operator):
            return

        for arg in node.args:
            self.visit(arg)

        if len(node.args) > 0:
            node.scale_factor = node.args[0].scale_factor

    def visit(self, node):

        if isinstance(node, (int, float, bool, str)):
            self._verbose_skip(node)
            return

        if isinstance(node, storage.variable):
            self._verbose_skip(node)
            return

        if isinstance(node, storage.placeholder):
            if node.value is not None:
                self._verbose_skip(node)
                return

            name = node.name

            if name in self.input_dict:
                # values are already assigned
                self._verbose_skip(node)
                return

            if name in self.input_generators:
                generator = self.input_generators[name]
                value = generator(node, self.num_samples)
                self.input_dict[name] = value

                if name not in self.input_scale_factors:
                    raise ValueError("scale_factor of '%s' not found." % name)
                node.scale_factor = self.input_scale_factors[name]

                self._verbose_node(node)
                return

            value = generate_samples(node, self.input_scale_factors,
                                     self.input_means, self.input_stds, self.num_samples)
            self.input_dict[name] = value

            if name not in self.input_scale_factors:
                raise ValueError("scale_factor of '%s' not found." % name)
            node.scale_factor = self.input_scale_factors[name]

            self._verbose_node(node)
            return

        if node.quantized:
            self._verbose_skip(node)
            return

        # constant
        if isinstance(node, storage.constant):
            q_value, scale_factor = util.quantize_linear(node.value, node.dtype.width)
            node.set_value(q_value)
            node.scale_factor = scale_factor
            node.quantized = True
            self._verbose_node(node)
            return

        # operators
        op_type = node.__class__.__name__

        if not _has_func(op_type) and isinstance(node, bt._Operator):
            self.generic_visit(node)
            node.quantized = True
            self._verbose_node(node)
            return

        node_func = _get_func(op_type)

        if node_func is None:
            raise NotImplementedError()

        node_func(self, node)
        node.quantized = True
        self._verbose_node(node)

    def _verbose_node(self, node):
        if self.verbose:
            print('[quantizer:done] {}'.format(str(node)))

    def _verbose_skip(self, node):
        if self.verbose:
            print('[quantizer:skip] {}'.format(str(node)))


def generate_samples(node, input_scale_factors, input_means, input_stds, num_samples):
    """
    Generate a dummy ndarray of random values to determine shift-amounts.

    Parameters
    ----------

    node : nngen.placeholder
        Target placeholder object to make a dummy ndarray.

    input_scale_factors: dict
        Input scale factor dictionary by name.

    input_means : dict
        Input mean dictionary by name

    input_stds : dict
        Input standard deviation dictionary by name

    num_samples : int
        Number of sampling trials to determine the right-shift amount.
        This value is multiplied to the batch size of the placeholder.
    """

    shape = list(node.shape)

    # enlarge the batch size by num_samples for shift-amount evaluation
    shape[0] *= num_samples
    shape = tuple(shape)
    length = np.multiply.reduce(shape)

    if node.name not in input_means:
        mean = np.array([0.0] * shape[-1]).astype(np.float32)
    else:
        mean = np.array(input_means[node.name]).astype(np.float32)

    if node.name not in input_stds:
        if node.name not in input_scale_factors:
            raise ValueError("scale_factor of '%s' not found." % node.name)
        scale_factor = input_scale_factors[node.name] * 1.0
        std = np.array([1.0] * shape[-1]).astype(np.float32) * scale_factor / 3.0
    else:
        std = np.array(input_stds[node.name]).astype(np.float32)

    # width = np.sqrt(12.0) * std / 2.0
    # v = np.random.uniform(-0.5, 0.5, size=length).reshape(shape) * width + mean
    v = np.random.normal(size=length).reshape(shape) * std + mean
    v = np.round(v).astype(np.int64)

    if node.dtype.signed:
        max_val = 2 ** (node.dtype.width - 1) - 1
        min_val = -1 * max_val
    else:
        max_val = 2 ** node.dtype.width - 1
        min_val = 0

    v = np.clip(v, min_val, max_val)

    return v


def quantize(outputs, input_scale_factors,
             input_means=None, input_stds=None, input_generators=None, num_samples=1,
             verbose=False):
    """
    Quantize pre-trained weights and determine right-shift amounts

    Parameters
    ----------

    outputs : list
        Output NNgen nodes.

    input_scale_factors: dict
        Input scale factor dictionary by name.

    input_means : dict
        Input mean dictionary by name

    input_stds : dict
        Input standard deviation dictionary by name

    input_generators : dict
        Dictionary by name for random sample value generator methods of placeholders.
        The methods must have arguments for node and num_samples.

    num_samples : int
        Number of sampling trials to determine the right-shift amount.
        This value is multiplied to the batch size of the placeholder.

    verbose : bool
        If True, quantization steps are printed out.
    """

    if isinstance(outputs, dict):
        outputs = outputs.values()

    if input_generators is None:
        input_generators = {}

    visitor = _QuantizeVisitor(input_scale_factors,
                               input_means, input_stds, input_generators, num_samples,
                               verbose)

    for output in outputs:
        visitor.visit(output)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections

import nngen.operator as operator

from . import util


def Pad(visitor, node):

    mode = 'constant'
    pads = []
    pad_value = 0.0

    all_pads_zero = True

    for attribute in node.attribute:
        if attribute.name == 'mode':
            mode = attribute.s.decode()

        if attribute.name == 'value':
            pad_value = attribute.f

        if attribute.name == 'pads':
            for pad in attribute.ints:
                pads.append(pad)
                if pad != 0:
                    all_pads_zero = False

    if all_pads_zero:
        node_name = util.get_name(node)
        src_name = node.input[0]
        src_op = visitor.visit(node.input[0])
        visitor.operators[node_name] = src_op
        return src_op

    if mode != 'constant':
        raise ValueError("Padding mode must be 'constant', not '%s'." % mode)

    if pad_value != 0.0:
        raise ValueError("Padding value must be a multiple of integer, not %f." % pad_value)

    pad_value = round(pad_value)

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]

    if len(input.shape) != 4:
        raise ValueError("not supported shape: %s" % str(tuple(shape)))

    # transpose data layout to nngen-compatible format
    input = util.transpose_layout(input, visitor.nngen_input_layout, visitor.onnx_input_layout)

    name = util.get_name(node)

    pads_pre = pads[:len(pads) // 2]
    pads_post = pads[len(pads) // 2:]

    pads_pre = [pads_pre[visitor.onnx_input_layout.index(l)]
                for l in visitor.nngen_input_layout]
    pads_post = [pads_post[visitor.onnx_input_layout.index(l)]
                 for l in visitor.nngen_input_layout]

    # padding layout: Top, Bottom, Left, Right
    padding = [pads_pre[1], pads_post[1], pads_pre[2], pads_post[2]]

    if name in visitor.value_dtypes:
        dtype = visitor.value_dtypes[name]
    else:
        dtype = visitor.default_operator_dtype

    kwargs = collections.OrderedDict()
    kwargs['padding'] = tuple(padding)
    kwargs['dtype'] = dtype
    kwargs['name'] = name

    c = operator.pad(input, **kwargs)
    c.layout = visitor.nngen_input_layout
    c.onnx_layout = visitor.onnx_input_layout

    return c



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


def slice_(value, begins, ends, strides,
           dtype=None, name=None, par=1,
           value_ram_size=None, out_ram_size=None,
           value_dtype=None):

    slices = to_slices(begins, ends, strides)
    return value[slices]


def to_slices(begins, ends, strides):
    slices = []

    for begin, end, stride in zip(begins, ends, strides):
        slices.append(slice(begin, end, stride))

    return tuple(slices)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from . import pool


def avg_pool_serial(value, ksize, stride, padding='SAME',
                    dtype=None, sum_dtype=None, name=None, par=1,
                    force_div=False,
                    value_ram_size=None, out_ram_size=None,
                    value_dtype=None):

    return pool.avg_pool(value, ksize, stride, padding,
                         dtype, sum_dtype, name, par,
                         force_div,
                         value_ram_size, out_ram_size,
                         value_dtype)


def max_pool_serial(value, ksize, stride, padding='SAME',
                    dtype=None, name=None, par=1,
                    value_ram_size=None, out_ram_size=None,
                    value_dtype=None):

    return pool.max_pool(value, ksize, stride, padding,
                         dtype, name, par,
                         value_ram_size, out_ram_size,
                         value_dtype)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


def eval(objs, **input_dict):
    memo = {}
    return [obj.eval(memo, input_dict) for obj in objs]



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math

import veriloggen as vg
import veriloggen.thread as vthread
import veriloggen.stream as vstream

from . import util

_tmp_counter = 0


def _tmp_name(prefix=None):
    global _tmp_counter
    if prefix is None:
        prefix = 'tmp'
    name = '%s_%d' % (prefix, _tmp_counter)
    _tmp_counter += 1
    return name


def mul(m, clk, rst,
        x_datawidth, x_point, x_signed,
        y_datawidth, y_point, y_signed,
        mul_width=None, mul_point=None, mul_signed=None):

    name = _tmp_name('mul')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)

    stream.sink(z, 'z')
    return stream


def mul_rshift(m, clk, rst,
               x_datawidth, x_point, x_signed,
               y_datawidth, y_point, y_signed,
               mul_width=None, mul_point=None, mul_signed=None):

    name = _tmp_name('mul')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.Sra(z, rshift)

    stream.sink(z, 'z')
    return stream


def mul_rshift_clip(m, clk, rst,
                    x_datawidth, x_point, x_signed,
                    y_datawidth, y_point, y_signed,
                    mul_width=None, mul_point=None, mul_signed=None,
                    out_width=None, out_point=None, out_signed=None,
                    asymmetric_clip=False):

    name = _tmp_name('mul_rshift_clip')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.Sra(z, rshift)

    clip_width = datawidth
    if mul_width is not None:
        clip_width = mul_width
    if out_width is not None:
        clip_width = out_width

    clip_signed = x_signed and y_signed
    if mul_signed is not None:
        clip_signed = mul_signed
    if out_signed is not None:
        clip_signed = out_signed

    clip_point = point
    if mul_point is not None:
        clip_point = mul_point
    if out_point is not None:
        clip_point = out_point

    p_th, n_th = util.clip_threshold(clip_width, clip_signed, asymmetric_clip)

    if clip_point > 0:
        p_th = p_th >> clip_point
        n_th = n_th >> clip_point
    elif clip_point < 0:
        p_th = p_th << -clip_point
        n_th = n_th << -clip_point

    p = stream.Mux(z > p_th, p_th, z)
    n = stream.Mux(z < n_th, n_th, z)
    z = stream.Mux(z >= 0, p, n)

    if out_width is not None:
        z.width = out_width
    if out_signed is not None:
        z.signed = out_signed
    if out_point is not None and z.point != out_point:
        z = stream.Cast(z, point=out_point)

    stream.sink(z, 'z')
    return stream


def mul_rshift_round(m, clk, rst,
                     x_datawidth, x_point, x_signed,
                     y_datawidth, y_point, y_signed,
                     mul_width=None, mul_point=None, mul_signed=None):

    name = _tmp_name('mul')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.SraRound(z, rshift)

    stream.sink(z, 'z')
    return stream


def mul_rshift_round_clip(m, clk, rst,
                          x_datawidth, x_point, x_signed,
                          y_datawidth, y_point, y_signed,
                          mul_width=None, mul_point=None, mul_signed=None,
                          out_width=None, out_point=None, out_signed=None,
                          asymmetric_clip=False):

    name = _tmp_name('mul_rshift_round_clip')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.SraRound(z, rshift)

    clip_width = datawidth
    if mul_width is not None:
        clip_width = mul_width
    if out_width is not None:
        clip_width = out_width

    clip_signed = x_signed and y_signed
    if mul_signed is not None:
        clip_signed = mul_signed
    if out_signed is not None:
        clip_signed = out_signed

    clip_point = point
    if mul_point is not None:
        clip_point = mul_point
    if out_point is not None:
        clip_point = out_point

    p_th, n_th = util.clip_threshold(clip_width, clip_signed, asymmetric_clip)

    if clip_point > 0:
        p_th = p_th >> clip_point
        n_th = n_th >> clip_point
    elif clip_point < 0:
        p_th = p_th << -clip_point
        n_th = n_th << -clip_point

    p = stream.Mux(z > p_th, p_th, z)
    n = stream.Mux(z < n_th, n_th, z)
    z = stream.Mux(z >= 0, p, n)

    if out_width is not None:
        z.width = out_width
    if out_signed is not None:
        z.signed = out_signed
    if out_point is not None and z.point != out_point:
        z = stream.Cast(z, point=out_point)

    stream.sink(z, 'z')
    return stream


def mul_rshift_round_madd(m, clk, rst,
                          x_datawidth, x_point, x_signed,
                          y_datawidth, y_point, y_signed,
                          mul_width=None, mul_point=None, mul_signed=None):

    name = _tmp_name('mul')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    frac = stream.Mux(rshift > 0, stream.Sll(1, rshift - 1), 0)
    frac.width = mul_width
    neg_frac = stream.Uminus(frac)
    neg_frac.width = datawidth
    neg_frac.latency = 0
    frac = stream.Mux(x >= 0, frac, neg_frac)
    frac.latency = 0
    frac.width = datawidth

    z = stream.Madd(x, y, frac)
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.Sra(z, rshift)

    stream.sink(z, 'z')
    return stream


def madd(m, clk, rst,
         x_datawidth, x_point, x_signed,
         y_datawidth, y_point, y_signed,
         z_datawidth, z_point, z_signed,
         sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('madd')
    datawidth = max(x_datawidth, y_datawidth, z_datawidth)
    point = max(x_point, y_point, z_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    z = stream.source('z', z_datawidth, z_point, z_signed)

    sum = stream.Madd(x, y, z)
    sum.latency = 4
    if sum_width is not None:
        sum.width = sum_width
    if sum_signed is not None:
        sum.signed = sum_signed
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    stream.sink(sum, 'sum')
    return stream


def madd_rshift(m, clk, rst,
                x_datawidth, x_point, x_signed,
                y_datawidth, y_point, y_signed,
                z_datawidth, z_point, z_signed,
                sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('madd')
    datawidth = max(x_datawidth, y_datawidth, z_datawidth)
    point = max(x_point, y_point, z_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    z = stream.source('z', z_datawidth, z_point, z_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    sum = stream.Madd(x, y, z)
    sum.latency = 4
    if sum_width is not None:
        sum.width = sum_width
    if sum_signed is not None:
        sum.signed = sum_signed
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)
    sum = stream.Sra(sum, rshift)

    stream.sink(sum, 'sum')
    return stream


def madd_rshift_clip(m, clk, rst,
                     x_datawidth, x_point, x_signed,
                     y_datawidth, y_point, y_signed,
                     z_datawidth, z_point, z_signed,
                     sum_width=None, sum_point=None, sum_signed=None,
                     out_width=None, out_point=None, out_signed=None,
                     asymmetric_clip=False):

    name = _tmp_name('madd')
    datawidth = max(x_datawidth, y_datawidth, z_datawidth)
    point = max(x_point, y_point, z_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    z = stream.source('z', z_datawidth, z_point, z_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    sum = stream.Madd(x, y, z)
    sum.latency = 4
    if sum_width is not None:
        sum.width = sum_width
    if sum_signed is not None:
        sum.signed = sum_signed
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)
    sum = stream.Sra(sum, rshift)

    clip_width = datawidth
    if sum_width is not None:
        clip_width = sum_width
    if out_width is not None:
        clip_width = out_width

    clip_signed = x_signed and y_signed and z_signed
    if sum_signed is not None:
        clip_signed = sum_signed
    if out_signed is not None:
        clip_signed = out_signed

    clip_point = point
    if mul_point is not None:
        clip_point = mul_point
    if out_point is not None:
        clip_point = out_point

    p_th, n_th = util.clip_threshold(clip_width, clip_signed, asymmetric_clip)

    if clip_point > 0:
        p_th = p_th >> clip_point
        n_th = n_th >> clip_point
    elif clip_point < 0:
        p_th = p_th << -clip_point
        n_th = n_th << -clip_point

    p = stream.Mux(sum > p_th, p_th, sum)
    n = stream.Mux(sum < n_th, n_th, sum)
    sum = stream.Mux(sum >= 0, p, n)

    if out_width is not None:
        sum.width = out_width
    if out_signed is not None:
        sum.signed = out_signed
    if out_point is not None and sum.point != out_point:
        sum = stream.Cast(sum, point=out_point)

    stream.sink(sum, 'sum')
    return stream


def mac(m, clk, rst,
        x_datawidth, x_point, x_signed,
        y_datawidth, y_point, y_signed,
        mul_width=None, mul_point=None, mul_signed=None,
        sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('mac')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    size = stream.parameter('size', signed=False)

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    sum, v = stream.ReduceAddValid(z, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def mac_rshift(m, clk, rst,
               x_datawidth, x_point, x_signed,
               y_datawidth, y_point, y_signed,
               mul_width=None, mul_point=None, mul_signed=None,
               sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('mac')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1
    size = stream.parameter('size', signed=False)

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.Sra(z, rshift)
    sum, v = stream.ReduceAddValid(z, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def mac_rshift_round(m, clk, rst,
                     x_datawidth, x_point, x_signed,
                     y_datawidth, y_point, y_signed,
                     mul_width=None, mul_point=None, mul_signed=None,
                     sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('mac')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1
    size = stream.parameter('size', signed=False)

    z = x * y
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.SraRound(z, rshift)
    sum, v = stream.ReduceAddValid(z, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def mac_rshift_round_madd(m, clk, rst,
                          x_datawidth, x_point, x_signed,
                          y_datawidth, y_point, y_signed,
                          mul_width=None, mul_point=None, mul_signed=None,
                          sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('mac')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1
    size = stream.parameter('size', signed=False)

    frac = stream.Mux(rshift > 0, stream.Sll(1, rshift - 1), 0)
    frac.width = mul_width

    z = stream.Madd(x, y, frac)
    z.latency = 4
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
    z = stream.Sra(z, rshift)
    sum, v = stream.ReduceAddValid(z, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def acc(m, clk, rst,
        datawidth, point, signed,
        sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('acc')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', datawidth, point, signed)
    size = stream.parameter('size', signed=False)

    sum, v = stream.ReduceAddValid(x, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def acc_rshift(m, clk, rst,
               datawidth, point, signed,
               sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('acc')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', datawidth, point, signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1
    size = stream.parameter('size', signed=False)

    sum, v = stream.ReduceAddValid(x, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    sum = stream.Sra(sum, rshift)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def acc_rshift_round(m, clk, rst,
                     datawidth, point, signed,
                     sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('acc')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', datawidth, point, signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1
    size = stream.parameter('size', signed=False)

    sum, v = stream.ReduceAddValid(x, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    sum = stream.SraRound(sum, rshift)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def acc_rshift_round_frac(m, clk, rst,
                          datawidth, point, signed,
                          sum_width=None, sum_point=None, sum_signed=None):

    name = _tmp_name('acc')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', datawidth, point, signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1
    size = stream.parameter('size', signed=False)

    frac = stream.Mux(rshift > 0, stream.Sll(1, rshift - 1), 0)
    frac.width = sum_width

    sum, v = stream.ReduceAddValid(x, size, width=sum_width, signed=sum_signed)
    if sum_point is not None and point != sum_point:
        sum = stream.Cast(sum, point=sum_point)

    sum = sum + frac
    sum = stream.Sra(sum, rshift)

    stream.sink(sum, 'sum')
    stream.sink(v, 'valid')
    return stream


def add_tree(m, clk, rst,
             datawidth, point, signed, num_vars):

    name = _tmp_name('add_tree')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    vars = [stream.source('var%d' % i, datawidth, point, signed)
            for i in range(num_vars)]

    if len(vars) == 1:
        sum = stream.Cast(vars[0])
    else:
        sum = Add3Tree(stream, *vars)
    stream.sink(sum, 'sum')
    return stream


def add_tree_rshift(m, clk, rst,
                    datawidth, point, signed, num_vars):

    name = _tmp_name('add_tree')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    vars = [stream.source('var%d' % i, datawidth, point, signed)
            for i in range(num_vars)]
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    sum = Add3Tree(stream, *vars)
    sum = stream.Sra(sum, rshift)
    stream.sink(sum, 'sum')
    return stream


def add_tree_rshift_round(m, clk, rst,
                          datawidth, point, signed, num_vars):

    name = _tmp_name('add_tree')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    vars = [stream.source('var%d' % i, datawidth, point, signed)
            for i in range(num_vars)]
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    sum = Add3Tree(stream, *vars)
    sum = stream.SraRound(sum, rshift)
    stream.sink(sum, 'sum')
    return stream


def add_tree_rshift_round_frac(m, clk, rst,
                               datawidth, point, signed, num_vars):

    name = _tmp_name('add_tree')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    vars = [stream.source('var%d' % i, datawidth, point, signed)
            for i in range(num_vars)]
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    frac = stream.Mux(rshift > 0, stream.Sll(1, rshift - 1), 0)
    frac.width = datawidth

    sum = Add3Tree(stream, *(vars + [frac]))
    sum = stream.Sra(sum, rshift)
    stream.sink(sum, 'sum')
    return stream


def _max(m, clk, rst,
         datawidth, point, signed, num_vars):

    name = _tmp_name('_max')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    vars = [stream.source('var%d' % i, datawidth, point, signed)
            for i in range(num_vars)]

    val = stream.Max(*vars)
    stream.sink(val, 'val')
    return stream


def average(m, clk, rst,
            datawidth, point, signed, num_vars):

    name = _tmp_name('average')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    vars = [stream.source('var%d' % i, datawidth, point, signed)
            for i in range(num_vars)]

    val = stream.Average(*vars)
    stream.sink(val, 'val')
    return stream


def div(m, clk, rst,
        x_datawidth, x_point, x_signed,
        y_datawidth, y_point, y_signed,
        div_width=None, div_point=None, div_signed=None):

    name = _tmp_name('div')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)

    z = stream.Div(x, y)
    if div_width is not None:
        z.width = div_width
    if div_signed is not None:
        z.signed = div_signed
    if div_point is not None and point != div_point:
        z = stream.Cast(z, point=div_point)

    stream.sink(z, 'z')
    return stream


def div_const(m, clk, rst,
              x_datawidth, x_point, x_signed,
              y_datawidth, y_point, y_signed,
              div_width=None, div_point=None, div_signed=None):

    name = _tmp_name('div_const')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)

    z = stream.Div(x, y)
    if div_width is not None:
        z.width = div_width
    if div_signed is not None:
        z.signed = div_signed
    if div_point is not None and point != div_point:
        z = stream.Cast(z, point=div_point)

    stream.sink(z, 'z')
    return stream


def div_const_frac(m, clk, rst,
                   x_datawidth, x_point, x_signed,
                   y_datawidth, y_point, y_signed,
                   div_width=None, div_point=None, div_signed=None):

    name = _tmp_name('div_const_frac')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)

    frac = stream.source('frac')
    frac.width = datawidth

    neg_frac = stream.Uminus(frac)
    neg_frac.width = datawidth
    neg_frac.latency = 0

    frac = stream.Mux(x >= 0, frac, neg_frac)
    frac.latency = 0
    frac.width = datawidth

    x_frac = stream.Add(x, frac)
    x_frac.latency = 0
    z = stream.Div(x_frac, y)
    if div_width is not None:
        z.width = div_width
    if div_signed is not None:
        z.signed = div_signed
    if div_point is not None and point != div_point:
        z = stream.Cast(z, point=div_point)

    stream.sink(z, 'z')
    return stream


def op3_tree(op, initval, latency, *args):
    if len(args) == 0:
        return initval

    if len(args) == 1:
        return args[0]

    if len(args) == 2:
        ret = op(args[0], args[1], initval)

        if latency is not None:
            ret.latency = latency

        return ret

    if len(args) == 3:
        ret = op(*args)

        if latency is not None:
            ret.latency = latency

        return ret

    log_num = int(math.ceil(math.log(len(args), 3))) - 1
    num = 3 ** log_num

    ret = op(op3_tree(op, initval, latency, *args[:num]),
             op3_tree(op, initval, latency, *args[num: num * 2]),
             op3_tree(op, initval, latency, *args[num * 2:]))

    if latency is not None:
        ret.latency = latency

    return ret


def Add3Tree(stream, *args):
    op = stream.AddN
    return op3_tree(op, vstream.Int(0, signed=True), None, *args)


def reduce_max(m, clk, rst,
               datawidth, point, signed):

    name = _tmp_name('_reduce_max')

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', datawidth, point, signed)
    size = stream.parameter('size', signed=False)

    if signed:
        initval = - 2 ** (datawidth - 1)
    else:
        initval = 0

    data, valid = stream.ReduceMaxValid(x, size, initval=initval,
                                        width=datawidth, signed=signed)

    stream.sink(data, 'data')
    stream.sink(valid, 'valid')

    return stream


def lshift_rshift(m, clk, rst,
                  x_datawidth, x_point, x_signed,
                  y_datawidth, y_point, y_signed,
                  mul_width=None, mul_point=None, mul_signed=None):

    if y_point != 0:
        raise ValueError('not supported')

    name = _tmp_name('lshift')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    abs_y = stream.Abs(y)
    sign_y = stream.Sign(y)

    z = stream.Sll(x, abs_y)
    z.latency = 0
    z = stream.Cast(z, signed=x_signed)
    z.latency = 0
    z = stream.Mux(sign_y, stream.Complement2(z), z)
    z.latency = 0
    z = stream.Cast(z, signed=x_signed)
    z.latency = 0
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
        z.latency = 0
    z = stream.SraRound(z, rshift)

    stream.sink(z, 'z')
    return stream


def updown_rshift(m, clk, rst,
                  x_datawidth, x_point, x_signed,
                  y_datawidth, y_point, y_signed,
                  mul_width=None, mul_point=None, mul_signed=None):

    if y_datawidth != 1:
        raise ValueError('not supported')

    if y_point != 0:
        raise ValueError('not supported')

    if y_signed:
        raise ValueError('not supported')

    name = _tmp_name('updown')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    z = stream.Mux(y, x, stream.Complement2(x))
    z.latency = 0
    z = stream.Cast(z, signed=x_signed)
    z.latency = 0
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
        z.latency = 0
    z = stream.SraRound(z, rshift)

    stream.sink(z, 'z')
    return stream


def updown_mask_rshift(m, clk, rst,
                       x_datawidth, x_point, x_signed,
                       y_datawidth, y_point, y_signed,
                       mul_width=None, mul_point=None, mul_signed=None):

    if y_datawidth != 2:
        raise ValueError('not supported')

    if y_point != 0:
        raise ValueError('not supported')

    if not y_signed:
        raise ValueError('not supported')

    name = _tmp_name('updown_mask')
    datawidth = max(x_datawidth, y_datawidth)
    point = max(x_point, y_point)

    stream = vthread.Stream(m, name, clk, rst, datawidth)
    x = stream.source('x', x_datawidth, x_point, x_signed)
    y = stream.source('y', y_datawidth, y_point, y_signed)
    rshift = stream.source('rshift', signed=False)
    rshift.width = int(math.ceil(math.log(datawidth, 2))) + 1

    z = stream.Mux(y > 0, x,
                   stream.Mux(y < 0, stream.Complement2(x), 0))
    z.latency = 0
    z = stream.Cast(z, signed=x_signed)
    z.latency = 0
    if mul_width is not None:
        z.width = mul_width
    if mul_signed is not None:
        z.signed = mul_signed
    if mul_point is not None and point != mul_point:
        z = stream.Cast(z, point=mul_point)
        z.latency = 0
    z = stream.SraRound(z, rshift)

    stream.sink(z, 'z')
    return stream
